import re

import cassiopeia.riotapi
import cassiopeia.type.core.common
import cassiopeia.type.dto.staticdata

try:
    from future.builtins.misc import super
except ImportError:
    pass


######################
# Champion Endpoints #
######################
@cassiopeia.type.core.common.inheritdocs
class SetItem(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.BlockItem

    def __str__(self):
        return "{item} ({count})".format(item=self.item, count=self.count)

    @property
    def count(self):
        """
        Returns:
            int: how many of this item are in the block
        """
        return self.data.count

    @property
    def item(self):
        """
        Returns:
            Item: the item itself
        """
        return cassiopeia.riotapi.get_item(self.data.id)


@cassiopeia.type.core.common.inheritdocs
class Item(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Item

    __stat_patterns = {
        "percent_base_attack_damage": "\\+(\\d+)% Base Attack Damage",
        "percent_cooldown_reduction": "\\+(\\d+) *% Cooldown Reduction",
        "armor_pen": "\\+(\\d+) <a href='FlatArmorPen'>Armor Penetration</a>",
        "percent_bonus_armor_pen": "\\+(\\d+)% <a href='BonusArmorPen'>Bonus Armor Penetration</a>",
        "magic_pen": "\\+(\\d+) <a href='FlatMagicPen'>Magic Penetration</a>",
        "percent_magic_pen": "\\+(\\d+)% <a href='TotalMagicPen'>Magic Penetration</a>",
        "gold_per_ten": "\\+(\\d+) *Gold per 10 seconds",
        "percent_ability_power": "Increases Ability Power by (\\d+)%",
        "life_steal": "(?:Dealing physical damage heals for (\\d+)% of the damage dealt)|(?:\\+(\\d+)% Life Steal)",
        "spell_vamp": "(?:\\+(\\d+)% <a href='SpellVamp'>Spell Vamp</a>)|(?:Your spells and abilities heal you for (\\d+)% of the damage dealt)",
        "percent_base_mana_regen": "\\+(\\d+)% Base Mana Regen",
        "percent_base_health_regen": "\\+(\\d+)% Base Health Regen",
        "percent_bonus_health": "\\+(\\d+)% Bonus Health",
        "percent_movespeed": "\\+(\\d+)% Movement Speed",
        "tenacity": "Tenacity:</unique> Reduces the duration of stuns, slows, taunts, fears, silences, blinds, polymorphs, and immobilizes by (\\d+)%"
    }

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def keywords(self):
        """
        Returns:
            str: a string formatted list of search keywords for this item in the shop
        """
        return self.data.colloq

    @property
    def consume_on_full(self):
        """
        Returns:
            bool: well, we don't know what this one is. let us know if you figure it out.
        """
        return self.data.consumeOnFull

    @property
    def consumable(self):
        """
        Returns:
            bool: whether the item is a consumable
        """
        return self.data.consumed

    @property
    def tier(self):
        """
        Returns:
            str: what tier the item is
        """
        return self.data.depth

    @property
    def description(self):
        """
        Returns:
            str: the description of the item
        """
        return self.data.description

    @property
    def effect(self):
        """
        Returns:
            dict<str, bool>: the item's effects
        """
        return self.data.effect

    @property
    def components(self):
        """
        Returns:
            list<Item>: the components for this item
        """
        return cassiopeia.riotapi.get_items([int(id_) for id_ in self.data.from_])

    @cassiopeia.type.core.common.lazyproperty
    def gold(self):
        """
        Returns:
            Gold: price information for this item
        """
        return Gold(self.data.gold) if self.data.gold else None

    @property
    def group(self):
        """
        Returns:
            str: the group for this item
        """
        return self.data.group

    @property
    def hide(self):
        """
        Returns:
            bool: well, we don't know what this one is. let us know if you figure it out.
        """
        return self.data.hide_from_all

    @property
    def id(self):
        """
        Returns:
            int: the ID of this item
        """
        return self.data.id

    @cassiopeia.type.core.common.lazyproperty
    def image(self):
        """
        Returns:
            Image: the image of this item
        """
        return Image(self.data.image) if self.data.image else None

    @property
    def in_store(self):
        """
        Returns:
            bool: well, we don't know what this one is. let us know if you figure it out.
        """
        return self.data.inStore

    @property
    def component_of(self):
        """
        Returns:
            list<Item>: the items this one is a component of
        """
        return cassiopeia.riotapi.get_items([int(id_) for id_ in self.data.into])

    @property
    def maps(self):
        """
        Returns:
            dict<Map, bool>: a listing of whether this item is available on each map
        """
        return {cassiopeia.type.core.common.Map(int(id_)): allowed for id_, allowed in self.data.maps.items()}

    @property
    def name(self):
        """
        Returns:
            str: the name of this item
        """
        return self.data.name

    @property
    def blurb(self):
        """
        Returns:
            str: the blurb for this item
        """
        return self.data.plaintext

    @property
    def required_champion(self):
        """
        Returns:
            Champion: the required champion for this item
        """
        return cassiopeia.riotapi.get_champion_by_name(self.data.requiredChampion) if self.data.requiredChampion else None

    @cassiopeia.type.core.common.lazyproperty
    def meta_data(self):
        """
        Returns:
            MetaData: meta data about this item
        """
        return MetaData(self.data.rune) if self.data.rune else None

    @property
    def sanitized_description(self):
        """
        Returns:
            str: the sanitized description of this item
        """
        return self.data.sanitizedDescription

    @property
    def special_recipe(self):
        """
        Returns:
            int: well, we don't know what this one is. let us know if you figure it out.
        """
        return self.data.specialRecipe

    @property
    def stacks(self):
        """
        Returns:
            int: the number of stacks this item can have
        """
        return self.data.stacks

    @cassiopeia.type.core.common.lazyproperty
    def stats(self):
        """
        Returns:
            ItemStats: the stats for this item
        """
        scraped_stats = {}
        for stat, regex in Item.__stat_patterns.items():
            match = re.search(regex, self.description)
            if match:
                value = [match.group(i) for i in range(1, re.compile(regex).groups + 1) if match.group(i) is not None]
                value = sum([float(v) for v in value])
                scraped_stats[stat] = value
        if scraped_stats and self.data.stats is None:
            self.data.stats = cassiopeia.type.dto.staticdata.BasicDataStats({})
        return ItemStats(self.data.stats, scraped_stats) if self.data.stats else None

    @property
    def tags(self):
        """
        Returns:
            list<str>: this item's tags for sorting items
        """
        return self.data.tags

    # Left out these: attack_range, percent_time_dead, percent_time_dead_per_level, time_dead, time_dead_per_level, gold_per_ten, starting_gold, percent_xp_bonus, xp_bonus
    __item_categories = {
        "Defense": {
            "Health": ["base_health", "health", "percent_health", "percent_bonus_health", "health_regen", "base_health_regen", "health_regen_per_level", "bonus_health_regen", "percent_health_regen", "percent_base_health_regen", "health_per_level", "base_health_regen", "bonus_health"],
            "Armor": ["armor", "bonus_armor", "armor_per_level", "base_armor", "percent_armor"],
            "Magic Resist": ["magic_resist_per_level", "base_magic_resist", "magic_resist", "percent_magic_resist", "bonus_magic_resist"],
            "Tenacity": ["tenacity"],
            "Other": ["dodge_chance", "dodge_chance_per_level", "block", "percent_block", "tenacity"]
        },
        "Attack": {
            "Damage": ["bonus_attack_damage", "attack_damage", "base_attack_damage", "percent_attack_damage", "percent_total_damage_increase", "attack_damage_per_level"],
            "Critical Strike": ["percent_critical_strike_damage", "critical_strike_chance_per_level", "critical_strike_damage", "critical_strike_damage_per_level", "critical_strike_chance", "critical_strike_chance_per_level"],
            "Attack Speed": ["percent_attack_speed", "attack_speed", "bonus_attack_speed", "attack_speed_per_level", "base_attack_speed", "percent_attack_speed_per_level"],
            "Life Steal": ["life_steal"],
            "Other": ["armor_penetration", "percent_armor_penetration", "percent_bonus_armor_penetration", "armor_penetration_per_level", "percent_armor_penetration_per_level", "magic_penetration", "percent_magic_penetration", "magic_penetration_per_level", "percent_magic_pen_per_level"]
        },
        "Magic": {
            "Ability Power": ["ability_power", "ability_power_per_level", "percent_ability_power"],
            "Cooldown Reduction": ["item_cooldown_reduction", "cooldown_reduction_per_level", "cooldown_reduction"],
            "Spell Vamp": ["spell_vamp"],
            "Mana": ["base_mana", "bonus_mana", "energy_per_level", "energy", "mana_per_level", "mana", "percent_mana", "base_mana_regen"],
            "Mana Regen": ["base_mana_regen", "bonus_mana_regen", "mana_regen_per_level", "percent_mana_regen", "percent_base_mana_regen", "mana_regen", "energy_regen", "energy_regen_per_level"],
            "Other": []
        },
        "Movement": {
            "Other Movement Items": ["out_of_combat_movespeed", "percent_movespeed", "movespeed_per_level", "movespeed", "percent_movespeed_per_level"],
            "Other": []
        }
    }
    __boot_ids = {"1001", "3117", "3158", "3111", "3006", "3009", "3020", "3047"}

    @cassiopeia.type.core.common.lazyproperty
    def categories(self):
        """
        Returns:
            list<str>: the shop categories that this item belongs to
        """
        if self.consumable:
            cats = {"Consumable"}
        else:
            cats = set()
        if self.stats is not None:
            for cat_name, cat in Item.__item_categories.items():
                for subcat, attrs in cat.items():
                    for attr in attrs:
                        if getattr(self.stats, attr, 0.0) != 0.0:
                            cats.add(cat_name)
                            cats.add(subcat)
        if str(self.id) in Item.__boot_ids or len(set(self.data.from_) & Item.__boot_ids) > 0:
            cats.add("Boots")
            cats.remove("Other Movement Items")

        return list(cats)


@cassiopeia.type.core.common.inheritdocs
class ItemSet(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Block

    def __str__(self):
        return "Item Set"

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)

    def __getitem__(self, index):
        return self.items[index]

    @cassiopeia.type.core.common.lazyproperty
    def items(self):
        """
        Returns:
            list<SetItem>: the items in this item set
        """
        return [SetItem(item) for item in self.data.items]

    @property
    def rec_math(self):
        """
        Returns:
            bool: well, we don't know what this one is. let us know if you figure it out.
        """
        return self.data.recMath

    @property
    def type(self):
        """
        Returns:
            str: what type the item set is (e.g. starting items)
        """
        return self.data.type


@cassiopeia.type.core.common.inheritdocs
class SpellVariables(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.SpellVars

    def __str__(self):
        return "Spell Variables"

    @property
    def coefficients(self):
        """
        Returns:
            list<float>: the coefficients for determining spell scaling
        """
        return self.data.coeff

    @property
    def dynamic(self):
        """
        Returns:
            str: whether the spell variables are dynamic
        """
        return self.data.dyn

    @property
    def key(self):
        """
        Returns:
            str: this spell's identifying key
        """
        return self.data.key

    @property
    def link(self):
        """
        Returns:
            str: the link to the image. See https://developer.riotgames.com/docs/static-data for more information.
        """
        return self.data.link

    @property
    def ranks_with(self):
        """
        Returns:
            str: what these variables rank with
        """
        return self.data.ranksWith


@cassiopeia.type.core.common.inheritdocs
class LevelTip(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.LevelTip

    def __str__(self):
        return "Level Tip"

    @property
    def effects(self):
        """
        Returns:
            list<list<float>>: the level-by-level replacements for {{ e# }} tags in other values
        """
        return self.data.effect

    @property
    def labels(self):
        """
        Returns:
            list<str>: the labels for the changes in effects
        """
        return self.data.label


@cassiopeia.type.core.common.inheritdocs
class ChampionStats(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Stats

    def __str__(self):
        return "Champion Stats"

    @property
    def armor(self):
        """
        Returns:
            float: armor
        """
        return self.data.armor

    @property
    def armor_per_level(self):
        """
        Returns:
            float: armor per level
        """
        return self.data.armorperlevel

    @property
    def attack_damage(self):
        """
        Returns:
            float: attack damage
        """
        return self.data.attackdamage

    @property
    def attack_damage_per_level(self):
        """
        Returns:
            float: attack damage per level
        """
        return self.data.attackdamageperlevel

    @property
    def attack_range(self):
        """
        Returns:
            float: attack range
        """
        return self.data.attackrange

    @property
    def attack_speed(self):
        """
        Returns:
            float: attack speed
        """
        return 0.625 / (1.0 + self.data.attackspeedoffset)

    @property
    def percent_attack_speed_per_level(self):
        """
        Returns:
            float: attack speed per level
        """
        return self.data.attackspeedperlevel / 100

    @property
    def critical_strike_chance(self):
        """
        Returns:
            float: crititical strike chance
        """
        return self.data.crit

    @property
    def critical_strike_chance_per_level(self):
        """
        Returns:
            float: crititical strike chance per level
        """
        return self.data.critperlevel

    @property
    def health(self):
        """
        Returns:
            float: health
        """
        return self.data.hp

    @property
    def health_per_level(self):
        """
        Returns:
            float: health per level
        """
        return self.data.hpperlevel

    @property
    def health_regen(self):
        """
        Returns:
            float: health regen
        """
        return self.data.hpregen

    @property
    def health_regen_per_level(self):
        """
        Returns:
            float: health regen per level
        """
        return self.data.hpregenperlevel

    @property
    def movespeed(self):
        """
        Returns:
            float: movespeed
        """
        return self.data.movespeed

    @property
    def mana(self):
        """
        Returns:
            float: mana
        """
        return self.data.mp

    @property
    def mana_per_level(self):
        """
        Returns:
            float: mana per level
        """
        return self.data.mpperlevel

    @property
    def mana_regen(self):
        """
        Returns:
            float: mana regen
        """
        return self.data.mpregen

    @property
    def mana_regen_per_level(self):
        """
        Returns:
            float: mana regen per level
        """
        return self.data.mpregenperlevel

    @property
    def magic_resist(self):
        """
        Returns:
            float: magic resist
        """
        return self.data.spellblock

    @property
    def magic_resist_per_level(self):
        """
        Returns:
            float: magic resist per level
        """
        return self.data.spellblockperlevel


@cassiopeia.type.core.common.inheritdocs
class Skin(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Skin

    def __init__(self, data, key):
        super().__init__(data)
        self.__key = key

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def splash(self):
        """
        Returns:
            str: the link to the splash art for this skin
        """
        return "http://ddragon.leagueoflegends.com/cdn/img/champion/splash/{0}_{1}.jpg".format(self.__key, self.number)

    @property
    def loading(self):
        """
        Returns:
            str: the link to the loading art for this skin
        """
        return "http://ddragon.leagueoflegends.com/cdn/img/champion/loading/{0}_{1}.jpg".format(self.__key, self.number)

    @property
    def id(self):
        """
        Returns:
            int: the ID of this skin
        """
        return self.data.id

    @property
    def name(self):
        """
        Returns:
            str: the name of this skin
        """
        return self.data.name

    @property
    def number(self):
        """
        Returns:
            int: where in the skin order this skin comes
        """
        return self.data.num


@cassiopeia.type.core.common.inheritdocs
class RecommendedItems(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Recommended

    def __str__(self):
        "Recommended Items for {champion} on {mode}".format(champion=self.data.champion, mode=self.mode)

    def __iter__(self):
        return iter(self.blocks)

    def __len__(self):
        return len(self.blocks)

    def __getitem__(self, index):
        return self.blocks[index]

    @cassiopeia.type.core.common.lazyproperty
    def item_sets(self):
        """
        Returns:
            list<ItemSet>: the sets of items that make up this reommended page
        """
        return [ItemSet(block) for block in self.data.blocks]

    @property
    def champion(self):
        """
        Returns:
            Champion: the champion these recommendations are for
        """
        return cassiopeia.riotapi.get_champion_by_name(self.data.champion) if self.data.champion else None

    @property
    def map(self):
        """
        Returns:
            str: the name of the map these recommendations are for
        """
        return self.data.map

    @property
    def mode(self):
        """
        Returns:
            GameMode: the game mode these recommendations are for
        """
        return cassiopeia.type.core.common.GameMode(self.data.mode) if self.data.mode else None

    @property
    def priority(self):
        """
        Returns:
            bool: whether this is a priority recommendation
        """
        return self.data.priority

    @property
    def name(self):
        """
        Returns:
            str: the name of this item
        """
        return self.data.title

    @property
    def type(self):
        """
        Returns:
            str: the type of recommended items
        """
        return self.data.type


@cassiopeia.type.core.common.inheritdocs
class Image(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Image

    def __str__(self):
        return "Image ({link})".format(link=self.link)

    @property
    def link(self):
        """
        Returns:
            str: the link to the image. See https://developer.riotgames.com/docs/static-data for more information.
        """
        return self.data.full

    @property
    def group(self):
        """
        Returns:
            int: how many of this item are in the block
        """
        return self.data.group

    @property
    def height(self):
        """
        Returns:
            int: the height of the image
        """
        return self.data.h

    @property
    def sprite(self):
        """
        Returns:
            str: the sprite image link. See https://developer.riotgames.com/docs/static-data for more information.
        """
        return self.data.sprite

    @property
    def width(self):
        """
        Returns:
            int: the width of the image
        """
        return self.data.w

    @property
    def x(self):
        """
        Returns:
            int: the x offset of the image
        """
        return self.data.x

    @property
    def y(self):
        """
        Returns:
            int: the y offset of the image
        """
        return self.data.y


@cassiopeia.type.core.common.inheritdocs
class Passive(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Passive

    def __str__(self):
        return self.name

    @property
    def description(self):
        """
        Returns:
            str: the description of the passive
        """
        return self.data.description

    @cassiopeia.type.core.common.lazyproperty
    def image(self):
        """
        Returns:
            Image: the image of this passive
        """
        return Image(self.data.image) if self.data.image else None

    @property
    def name(self):
        """
        Returns:
            str: the name of this passive
        """
        return self.data.name

    @property
    def sanitized_description(self):
        """
        Returns:
            str: the sanitized description of this passive
        """
        return self.data.sanitizedDescription


@cassiopeia.type.core.common.inheritdocs
class ChampionInfo(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Info

    def __str__(self):
        return "Champion Information"

    @property
    def physical(self):
        """
        Returns:
            int: physical damage output rating (out of 10)
        """
        return self.data.attack

    @property
    def defense(self):
        """
        Returns:
            int: defensive rating (out of 10)
        """
        return self.data.defense

    @property
    def difficulty(self):
        """
        Returns:
            int: difficulty rating (out of 10)
        """
        return self.data.difficulty

    @property
    def magic(self):
        """
        Returns:
            int: magic damage output rating (out of 10)
        """
        return self.data.magic


@cassiopeia.type.core.common.inheritdocs
class Spell(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.ChampionSpell

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.key == other.key

    def __ne__(self, other):
        return self.key != other.key

    def __hash__(self):
        return hash(self.key)

    @cassiopeia.type.core.common.lazyproperty
    def alternate_images(self):
        """
        Returns:
            list<Image>: the alternate images for this spell
        """
        return [Image(img) for img in self.data.altimages]

    @property
    def cooldowns(self):
        """
        Returns:
            list<float>: the cooldowns of this spell level-by-level
        """
        return self.data.cooldown

    @property
    def cooldown_burn(self):
        """
        Returns:
            str: a string formatted list of the spell's cooldowns by level
        """
        return self.data.cooldownBurn

    @property
    def costs(self):
        """
        Returns:
            list<int>: the cost of the spell level-by-level
        """
        return self.data.cost

    @property
    def cost_burn(self):
        """
        Returns:
            str: a string formatted list of the spell's cost by level
        """
        return self.data.costBurn

    @property
    def cost_type(self):
        """
        Returns:
            str: what the spell costs to use (e.g. mana or energy)
        """
        return self.data.costType

    @property
    def description(self):
        """
        Returns:
            str: the description of the spell
        """
        return self.data.description

    @property
    def effects(self):
        """
        Returns:
            list<list<float>>: the level-by-level replacements for {{ e# }} tags in other values
        """
        return self.data.effect

    @property
    def effect_burn(self):
        """
        Returns:
            list<str>: the string formatted replacements for {{ e# }} tags in other values by level
        """
        return self.data.effectBurn

    @cassiopeia.type.core.common.lazyproperty
    def image(self):
        """
        Returns:
            Image: the image of this spell
        """
        return Image(self.data.image) if self.data.image else None

    @property
    def key(self):
        """
        Returns:
            str: this spell's identifying key
        """
        return self.data.key

    @cassiopeia.type.core.common.lazyproperty
    def level_tip(self):
        """
        Returns:
            LevelTip: the level-up tips for this spell
        """
        return LevelTip(self.data.leveltip) if self.data.leveltip else None

    @property
    def max_rank(self):
        """
        Returns:
            int: the maximum level for this spell
        """
        return self.data.maxrank

    @property
    def name(self):
        """
        Returns:
            str: the name of this spell
        """
        return self.data.name

    @property
    def range(self):
        """
        Returns:
            self | list<int>: the level-by-level range of this spell
        """
        return self.data.range

    @property
    def range_burn(self):
        """
        Returns:
            str: the string formatted range of this spell by level
        """
        return self.data.rangeBurn

    @property
    def resource(self):
        """
        Returns:
            str: what resource this spell uses
        """
        return self.data.resource

    @property
    def sanitized_description(self):
        """
        Returns:
            str: the sanitized description of this spell
        """
        return self.data.sanitizedDescription

    @property
    def sanitized_tooltip(self):
        """
        Returns:
            str: the sanitized tooltip for this spell
        """
        return self.data.sanitizedTooltip

    @property
    def tooltip(self):
        """
        Returns:
            str: the tooltip for this spell
        """
        return self.data.tooltip

    @cassiopeia.type.core.common.lazyproperty
    def variables(self):
        """
        Returns:
            SpellVariables: the variables that determine this spell's damage
        """
        return [SpellVariables(svars) for svars in self.data.vars]

    def __replace_variables(self, text, level, rank):
        if level < 1 or level > 18:
            raise ValueError("Not a valid champion level")
        if rank < 1 or rank > self.max_rank:
            raise ValueError("Not a valid spell rank")

        i = 1
        for effect in self.effects:
            if effect:
                text = text.replace("{{{{ e{i} }}}}".format(i=i), str(effect[rank - 1]))
                i += 1

        for svar in self.variables:
            val = svar.coefficients[0]
            if len(svar.coefficients) == self.max_rank:
                val = svar.coefficients[rank - 1]
            elif svar.coefficients == 18:
                val = svar.coefficients[level - 1]
            replacement = str(val)

            if svar.link == "attackdamage":
                replacement = replacement + " AD"
            elif svar.link == "spelldamage":
                replacement = replacement + " AP"

            text = text.replace("{{{{ {key} }}}}".format(key=svar.key), replacement)

        return text

    @cassiopeia.type.core.common.immutablemethod
    def tooltip_for_level(self, level, rank):
        """Gets the tooltip for this spell for a specific level/rank

        Args:
            level (int): the level of the champion
            rank (int): the rank of this spell

        Returns:
            int: how many of this item are in the block
        """
        return self.__replace_variables(self.tooltip, level, rank)

    @cassiopeia.type.core.common.immutablemethod
    def sanitized_tooltip_for_level(self, level, rank):
        """Gets the sanitized tooltip for this spell for a specific level/rank

        Args:
            level (int): the level of the champion
            rank (int): the rank of this spell

        Returns:
            str: the sanitized tooltip for this spell for the specified level and rank
        """
        return self.__replace_variables(self.sanitized_tooltip, level, rank)


@cassiopeia.type.core.common.inheritdocs
class Champion(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Champion

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def ally_tips(self):
        """
        Returns:
            list<str>: the set of tips for allies of this champion
        """
        return self.data.allytips

    @property
    def blurb(self):
        """
        Returns:
            list<SetItem>: the items in this item set
        """
        return self.data.blurb

    @property
    def enemy_tips(self):
        """
        Returns:
            list<str>: the set of tips for enemies of this champion
        """
        return self.data.enemytips

    @property
    def id(self):
        """
        Returns:
            int: the ID of this champion
        """
        return self.data.id

    @cassiopeia.type.core.common.lazyproperty
    def image(self):
        """
        Returns:
            Image: the image of this champion
        """
        return Image(self.data.image) if self.data.image else None

    @cassiopeia.type.core.common.lazyproperty
    def info(self):
        """
        Returns:
            ChampionInfo: ratings of what this champion is good/bad at
        """
        return ChampionInfo(self.data.info) if self.data.info else None

    @property
    def key(self):
        """
        Returns:
            str: this champion's identifying key
        """
        return self.data.key

    @property
    def lore(self):
        """
        Returns:
            str: this champion's lore
        """
        return self.data.lore

    @property
    def name(self):
        """
        Returns:
            str: the name of this champion
        """
        return self.data.name

    @property
    def resource(self):
        """
        Returns:
            list<SetItem>: the items in this item set
        """
        return self.data.partype

    @cassiopeia.type.core.common.lazyproperty
    def passive(self):
        """
        Returns:
            Passive: this champion's passive
        """
        return Passive(self.data.passive) if self.data.passive else None

    @cassiopeia.type.core.common.lazyproperty
    def recommended_items(self):
        """
        Returns:
            list<RecommendedItems>: item recommendations for this champion
        """
        return [RecommendedItems(rec) for rec in self.data.recommended]

    @cassiopeia.type.core.common.lazyproperty
    def skins(self):
        """
        Returns:
            list<Skin>: this champion's skins
        """
        return [Skin(skin, self.key) for skin in self.data.skins]

    @cassiopeia.type.core.common.lazyproperty
    def spells(self):
        """
        Returns:
            list<Spell>: this champion's spells
        """
        return [Spell(spell) for spell in self.data.spells]

    @cassiopeia.type.core.common.lazyproperty
    def stats(self):
        """
        Returns:
            ChampionStats: the stats for this champion
        """
        return ChampionStats(self.data.stats) if self.data.stats else None

    @property
    def tags(self):
        """
        Returns:
            list<str>: this champions's tags for sorting champions
        """
        return self.data.tags

    @property
    def title(self):
        """
        Returns:
            str: this champion's title
        """
        return self.data.title

    @cassiopeia.type.core.common.immutablemethod
    def status(self):
        """
        Returns:
            list<SetItem>: the items in this item set
        """
        return cassiopeia.riotapi.get_champion_status(self)

    @cassiopeia.type.core.common.immutablemethod
    def mastery_level(self, summoner):
        """Gets the ChampionMastery object for the specified summoner
        Args:
            summoner (Summoner): the summoner to get champion mastery for
        Returns:
            bool: well, we don't know what this one is. let us know if you figure it out.
        """
        return cassiopeia.riotapi.get_champion_mastery(summoner, self)


##################
# Item Endpoints #
##################
@cassiopeia.type.core.common.inheritdocs
class MetaData(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.MetaData

    def __str__(self):
        return "Meta Data"

    @property
    def rune(self):
        """
        Returns:
            bool: whether the item is a rune
        """
        return self.data.isRune

    @property
    def tier(self):
        """
        Returns:
            str: what tier the item is
        """
        return self.data.tier

    @property
    def type(self):
        """
        Returns:
            str: the type of meta data
        """
        return self.data.type


@cassiopeia.type.core.common.inheritdocs
class Gold(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Gold

    def __str__(self):
        return "{price} gold".format(price=self.total)

    @property
    def base(self):
        """
        Returns:
            int: the base price of the item
        """
        return self.data.base

    @property
    def purchasable(self):
        """
        Returns:
            bool: whether the item can be bought
        """
        return self.data.purchasable

    @property
    def sell(self):
        """
        Returns:
            int: the sell price of the item
        """
        return self.data.sell

    @property
    def total(self):
        """
        Returns:
            int: the total price of the item
        """
        return self.data.total


@cassiopeia.type.core.common.inheritdocs
class ItemStats(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.BasicDataStats

    __stats = set()
    for tag, tags in Item._Item__item_categories.items():
        for tag, stats in tags.items():
            for stat in stats:
                __stats.add(stat)

    def __init__(self, data, scraped_stats={}):
        super().__init__(data)
        for k, v in scraped_stats.items():
            if ("percent" in k or "spell_vamp" in k or "life_steal" in k or "tenacity" in k) and v > 1.0:
                scraped_stats[k] = v / 100.0
        self.__scraped_stats = scraped_stats

    def __str__(self):
        stats = {}
        for stat in ItemStats._ItemStats__stats:
            value = getattr(self, stat, 0.0)
            if value != 0.0:
                stats[stat] = value
        return str(stats)

    @property
    def armor(self):
        """
        Returns:
            float: armor
        """
        return self.data.FlatArmorMod

    @property
    def attack_speed(self):
        """
        Returns:
            float: attack speed
        """
        return self.data.FlatAttackSpeedMod

    @property
    def block(self):
        """
        Returns:
            float: blocked damage per attack
        """
        return self.data.FlatBlockMod

    @property
    def percent_block(self):
        """
        Returns:
            float: percent blocked damage per attack
        """
        return self.data.PercentBlockMod

    @property
    def critical_strike_chance(self):
        """
        Returns:
            float: critical strike chance
        """
        return self.data.FlatCritChanceMod + self.data.PercentCritChanceMod

    @property
    def critical_strike_damage(self):
        """
        Returns:
            float: critical strike damage modification
        """
        return self.data.FlatCritDamageMod

    @property
    def xp_bonus(self):
        """
        Returns:
            float: experience bonus
        """
        return self.data.FlatEXPBonus

    @property
    def energy(self):
        """
        Returns:
            float: energy
        """
        return self.data.FlatEnergyPoolMod

    @property
    def energy_regen(self):
        """
        Returns:
            float: energy regen
        """
        return self.data.FlatEnergyRegenMod

    @property
    def health(self):
        """
        Returns:
            flaot: health
        """
        return self.data.FlatHPPoolMod

    @property
    def health_regen(self):
        """
        Returns:
            float: health regen
        """
        return self.data.FlatHPRegenMod

    @property
    def mana(self):
        """
        Returns:
            float: mana
        """
        return self.data.FlatMPPoolMod

    @property
    def mana_regen(self):
        """
        Returns:
            float mana regen
        """
        return self.data.FlatMPRegenMod

    @property
    def ability_power(self):
        """
        Returns:
            float: ability power
        """
        return self.data.FlatMagicDamageMod

    @property
    def movespeed(self):
        """
        Returns:
            float: movespeed
        """
        return self.data.FlatMovementSpeedMod

    @property
    def attack_damage(self):
        """
        Returns:
            float: attack damage
        """
        return self.data.FlatPhysicalDamageMod

    @property
    def magic_resist(self):
        """
        Returns:
            float: magic resist
        """
        return self.data.FlatSpellBlockMod

    @property
    def tenacity(self):
        """
        Returns:
            float: tenacity
        """
        return self.__scraped_stats.get("tenacity", 0.0)

    @property
    def percent_armor(self):
        """
        Returns:
            float: percent armor
        """
        return self.data.PercentArmorMod

    @property
    def percent_attack_speed(self):
        """
        Returns:
            float: percent attack speed
        """
        return self.data.PercentAttackSpeedMod

    @property
    def percent_critical_strike_damage(self):
        """
        Returns:
            float: percent critical strike damage modification
        """
        return self.data.PercentCritDamageMod

    @property
    def percent_xp_bonus(self):
        """
        Returns:
            float: percent experience bonus
        """
        return self.data.PercentEXPBonus

    @property
    def percent_health(self):
        """
        Returns:
            float: percent health
        """
        return self.data.PercentHPPoolMod

    @property
    def percent_bonus_health(self):
        """
        Returns:
            float: percent bonus health
        """
        return self.__scraped_stats.get("percent_bonus_health", 0.0)

    @property
    def percent_health_regen(self):
        """
        Returns:
            float: percent health regen
        """
        return self.data.PercentHPRegenMod

    @property
    def percent_base_health_regen(self):
        """
        Returns:
            float: percent base health regen
        """
        return self.__scraped_stats.get("percent_base_health_regen", 0.0)

    @property
    def life_steal(self):
        """
        Returns:
            float: life steal
        """
        return self.data.PercentLifeStealMod + self.__scraped_stats.get("life_steal", 0.0)

    @property
    def percent_mana(self):
        """
        Returns:
            float: percent mana
        """
        return self.data.PercentMPPoolMod

    @property
    def percent_mana_regen(self):
        """
        Returns:
            float: percent mana regen
        """
        return self.data.PercentMPRegenMod

    @property
    def percent_base_mana_regen(self):
        """
        Returns:
            float: percent base mana regen
        """
        return self.__scraped_stats.get("base_mana_regen", 0.0) + self.__scraped_stats.get("percent_base_mana_regen", 0.0)

    @property
    def percent_ability_power(self):
        """
        Returns:
            float: percent ability power
        """
        return self.data.PercentMagicDamageMod + self.__scraped_stats.get("percent_ability_power", 0.0)

    @property
    def percent_movespeed(self):
        """
        Returns:
            float: percent movespeed
        """
        return self.data.PercentMovementSpeedMod + self.__scraped_stats.get("percent_movespeed", 0.0)

    @property
    def percent_attack_damage(self):
        """
        Returns:
            float: percent attack damage
        """
        return self.data.PercentPhysicalDamageMod

    @property
    def percent_base_attack_damage(self):
        """
        Returns:
            float: percent attack damage
        """
        return self.__scraped_stats.get("percent_base_attack_damage", 0.0)

    @property
    def percent_magic_resist(self):
        """
        Returns:
            float: percent magic resist
        """
        return self.data.PercentSpellBlockMod

    @property
    def spell_vamp(self):
        """
        Returns:
            float: spell vamp
        """
        return self.data.PercentSpellVampMod + self.__scraped_stats.get("spell_vamp", 0.0)

    @property
    def armor_per_level(self):
        """
        Returns:
            float: armor per level
        """
        return self.data.rFlatArmorModPerLevel

    @property
    def armor_penetration(self):
        """
        Returns:
            float: armor penetration
        """
        return self.data.rFlatArmorPenetrationMod + self.__scraped_stats.get("armor_pen", 0.0)

    @property
    def armor_penetration_per_level(self):
        """
        Returns:
            float: armor penetration per level
        """
        return self.data.rFlatArmorPenetrationModPerLevel

    @property
    def critical_strike_chance_per_level(self):
        """
        Returns:
            float: critical strike chance per level
        """
        return self.data.rFlatCritChanceModPerLevel

    @property
    def critical_strike_damage_per_level(self):
        """
        Returns:
            float: critical strike damage per level
        """
        return self.data.rFlatCritDamageModPerLevel

    @property
    def dodge_chance(self):
        """
        Returns:
            float: dodge chance
        """
        return self.data.rFlatDodgeMod + self.data.PercentDodgeMod

    @property
    def dodge_chance_per_level(self):
        """
        Returns:
            float: dodge change per level
        """
        return self.data.rFlatDodgeModPerLevel

    @property
    def energy_per_level(self):
        """
        Returns:
            float: energy per level
        """
        return self.data.rFlatEnergyModPerLevel

    @property
    def energy_regen_per_level(self):
        """
        Returns:
            float: energy regen per level
        """
        return self.data.rFlatEnergyRegenModPerLevel

    @property
    def gold_per_ten(self):
        """
        Returns:
            float: gold per 10 seconds
        """
        return self.data.rFlatGoldPer10Mod + self.__scraped_stats.get("gold_per_ten", 0.0)

    @property
    def health_per_level(self):
        """
        Returns:
            float: health per level
        """
        return self.data.rFlatHPModPerLevel

    @property
    def health_regen_per_level(self):
        """
        Returns:
            float: health regen per level
        """
        return self.data.rFlatHPRegenModPerLevel

    @property
    def mana_per_level(self):
        """
        Returns:
            float: armor per level
        """
        return self.data.rFlatMPModPerLevel

    @property
    def mana_regen_per_level(self):
        """
        Returns:
            float: mana regen per level
        """
        return self.data.rFlatMPRegenModPerLevel

    @property
    def ability_power_per_level(self):
        """
        Returns:
            float: ability power per level
        """
        return self.data.rFlatMagicDamageModPerLevel

    @property
    def magic_penetration(self):
        """
        Returns:
            float: magic penetration
        """
        return self.data.rFlatMagicPenetrationMod + self.__scraped_stats.get("magic_pen", 0.0)

    @property
    def magic_penetration_per_level(self):
        """
        Returns:
            float: magic penetration per level
        """
        return self.data.rFlatMagicPenetrationModPerLevel

    @property
    def movespeed_per_level(self):
        """
        Returns:
            float: movespeed per level
        """
        return self.data.rFlatMovementSpeedModPerLevel

    @property
    def attack_damage_per_level(self):
        """
        Returns:
            float attack damage per level
        """
        return self.data.rFlatPhysicalDamageModPerLevel

    @property
    def magic_resist_per_level(self):
        """
        Returns:
            float: magic resist per level
        """
        return self.data.rFlatSpellBlockModPerLevel

    @property
    def time_dead(self):
        """
        Returns:
            float: time dead modification
        """
        return abs(self.data.rFlatTimeDeadMod)

    @property
    def time_dead_per_level(self):
        """
        Returns:
            float: time dead modification per level
        """
        return abs(self.data.rFlatTimeDeadModPerLevel)

    @property
    def percent_armor_penetration(self):
        """
        Returns:
            float: percent armor pentration
        """
        return self.data.rPercentArmorPenetrationMod + self.__scraped_stats.get("percent_armor_pen", 0.0)

    @property
    def percent_armor_penetration_per_level(self):
        """
        Returns:
            float: percent armor penetration per level
        """
        return self.data.rPercentArmorPenetrationModPerLevel

    @property
    def percent_bonus_armor_penetration(self):
        """
        Returns:
            float: percent armor pentration
        """
        return self.__scraped_stats.get("percent_bonus_armor_pen", 0.0)

    @property
    def percent_attack_speed_per_level(self):
        """
        Returns:
            float: percent attack speed per level
        """
        return self.data.rPercentAttackSpeedModPerLevel

    @property
    def cooldown_reduction(self):
        """
        Returns:
            float: cooldown reduction
        """
        return abs(self.data.rPercentCooldownMod + self.__scraped_stats.get("percent_cooldown_reduction", 0.0))

    @property
    def cooldown_reduction_per_level(self):
        """
        Returns:
            float: cooldown reduction per level
        """
        return abs(self.data.rPercentCooldownModPerLevel)

    @property
    def percent_magic_penetration(self):
        """
        Returns:
            float: percent magic penetration
        """
        return self.data.rPercentMagicPenetrationMod + self.__scraped_stats.get("percent_magic_pen", 0.0)

    @property
    def percent_magic_pen_per_level(self):
        """
        Returns:
            float: percent magic penetration per level
        """
        return self.data.rPercentMagicPenetrationModPerLevel

    @property
    def percent_movespeed_per_level(self):
        """
        Returns:
            float: percent movespeed per level
        """
        return self.data.rPercentMovementSpeedModPerLevel

    @property
    def percent_time_dead(self):
        """
        Returns:
            float: percent time dead modification
        """
        return abs(self.data.rPercentTimeDeadMod)

    @property
    def percent_time_dead_per_level(self):
        """
        Returns:
            float: percent time dead modification per level
        """
        return abs(self.data.rPercentTimeDeadModPerLevel)


################
# Map Endpoint #
################
@cassiopeia.type.core.common.inheritdocs
class MapDetails(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.MapDetails

    def __str__(self):
        return self.data.name

    def __eq__(self, other):
        return self.data.mapId == other.data.mapId

    def __ne__(self, other):
        return self.data.mapId != other.data.mapId

    def __hash__(self):
        return hash(self.data.mapId)

    @cassiopeia.type.core.common.lazyproperty
    def image(self):
        """
        Returns:
            Image: the image of this map
        """
        return Image(self.data.image) if self.data.image else None

    @property
    def map_id(self):
        """
        Returns:
            int: the ID of this map
        """
        return self.data.mapId

    @property
    def map(self):
        """
        Returns:
            Map: the type of this map
        """
        return cassiopeia.type.core.common.Map(self.data.mapId) if self.data.mapId else None

    @cassiopeia.type.core.common.lazyproperty
    def unpurchasable_items(self):
        """
        Returns:
            list<Item>: the items that can't be bought on this map
        """
        return list(filter(None, cassiopeia.riotapi.get_items(self.data.unpurchasableItemList)))


#####################
# Mastery Endpoints #
#####################
@cassiopeia.type.core.common.inheritdocs
class Mastery(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Mastery

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def descriptions(self):
        """
        Returns:
            list<str>: descriptions of this mastery by rank
        """
        return self.data.description

    @property
    def id(self):
        """
        Returns:
            int: the ID of this mastery
        """
        return self.data.id

    @cassiopeia.type.core.common.lazyproperty
    def image(self):
        """
        Returns:
            Image: the image of this mastery
        """
        return Image(self.data.image) if self.data.image else None

    @property
    def tree(self):
        """
        Returns:
            str: which mastery tree this mastery belongs to
        """
        return cassiopeia.type.core.common.MasteryType(self.data.masteryTree)

    @property
    def name(self):
        """
        Returns:
            str: the name of this mastery
        """
        return self.data.name.strip()

    @property
    def prerequisite(self):
        """
        Returns:
            Mastery: the prerequisite for this mastery
        """
        return cassiopeia.riotapi.get_mastery(int(self.data.prereq)) if int(self.data.prereq) else None

    @property
    def max_rank(self):
        """
        Returns:
            int: the max rank for this mastery
        """
        return self.data.ranks

    @property
    def sanitized_descriptions(self):
        """
        Returns:
            list<str>: sanitized descriptions of this mastery by rank
        """
        return self.data.sanitizedDescription


##################
# Realm Endpoint #
##################
@cassiopeia.type.core.common.inheritdocs
class Realm(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Realm

    def __str__(self):
        return "Realm"

    def __eq__(self, other):
        return self.cdn == other.cdn

    def __ne__(self, other):
        return self.cdn != other.cdn

    def __hash__(self):
        return hash(self.cdn)

    @property
    def cdn(self):
        """
        Returns:
            str: the base CDN url
        """
        return self.data.cdn

    @property
    def css(self):
        """
        Returns:
            str: the latest version of the Dragon Magic's css file
        """
        return self.data.css

    @property
    def dragon_magic(self):
        """
        Returns:
            str: the lastest version of Dragon Magic
        """
        return self.data.dd

    @property
    def language(self):
        """
        Returns:
            str: the default locale for this realm
        """
        return self.data.l

    @property
    def legacy(self):
        """
        Returns:
            str: the legacy script mode for IE6 or older
        """
        return self.data.lg

    @property
    def data_type_versions(self):
        """
        Returns:
            dict<str, str>: the latest versions for listed data types
        """
        return self.data.n

    @property
    def profile_icon_id_max(self):
        """
        Returns:
            int: the largest profileicon id that can be used under 500.0 Any profileicon that is requested between this number and 500 should be mapped to 0.0.
        """
        return self.data.profileiconmax

    @property
    def store(self):
        """
        Returns:
            str: additional api data drawn from other sources that may be related to data dragon functionality
        """
        return self.data.store

    @property
    def version(self):
        """
        Returns:
            str: the current version of this file
        """
        return self.data.v


##################
# Rune Endpoints #
##################
@cassiopeia.type.core.common.inheritdocs
class Rune(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.Rune

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def description(self):
        """
        Returns:
            str: the description of the rune
        """
        return self.data.description

    @property
    def id(self):
        """
        Returns:
            int: the ID of this rune
        """
        return self.data.id

    @cassiopeia.type.core.common.lazyproperty
    def image(self):
        """
        Returns:
            Image: the image of this rune
        """
        return Image(self.data.image) if self.data.image else None

    @property
    def name(self):
        """
        Returns:
            str: the name of this rune
        """
        return self.data.name

    @cassiopeia.type.core.common.lazyproperty
    def meta_data(self):
        """
        Returns:
            MetaData: meta data about this rune
        """
        return MetaData(self.data.rune) if self.data.rune else None

    @property
    def sanitized_description(self):
        """
        Returns:
            str: the sanitized description of this rune
        """
        return self.data.sanitizedDescription

    @cassiopeia.type.core.common.lazyproperty
    def stats(self):
        """
        Returns:
            ItemStats: the stats for this rune
        """
        return ItemStats(self.data.stats) if self.data.stats else None

    @property
    def tags(self):
        """
        Returns:
            list<str>: this rune's tags for sorting runes
        """
        return self.data.tags

    @property
    def rune_type(self):
        """
        Returns:
            str: what type of rune this is
        """
        try:
            return next(iter(set(self.tags).intersection({"mark", "seal", "glyph", "quintessence"})))
        except:
            return ""


############################
# Summoner Spell Endpoints #
############################
@cassiopeia.type.core.common.inheritdocs
class SummonerSpell(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.staticdata.SummonerSpell

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def cooldowns(self):
        """
        Returns:
            list<float>: the cooldowns of this spell level-by-level
        """
        return self.data.cooldown

    @property
    def cooldown_burn(self):
        """
        Returns:
            str: formatted list of the spell's cooldowns by level
        """
        return self.data.cooldownBurn

    @property
    def costs(self):
        """
        Returns:
            list<int>: the cost of the spell level-by-level
        """
        return self.data.cost

    @property
    def cost_burn(self):
        """
        Returns:
            str: formatted list of the spell's cost by level
        """
        return self.data.costBurn

    @property
    def cost_type(self):
        """
        Returns:
            str: what the spell costs to use (e.g. mana or energy)
        """
        return self.data.costType

    @property
    def description(self):
        """
        Returns:
            str: the description of the summoner spell
        """
        return self.data.description

    @property
    def effects(self):
        """
        Returns:
            list<list<float>>: the level-by-level replacements for {{ e# }} tags in other values
        """
        return self.data.effect

    @property
    def effect_burn(self):
        """
        Returns:
            list<str>: the string formatted replacements for {{ e# }} tags in other values by level
        """
        return self.data.effectBurn

    @property
    def id(self):
        """
        Returns:
            int: the ID of this summoner spell
        """
        return self.data.id

    @cassiopeia.type.core.common.lazyproperty
    def image(self):
        """
        Returns:
            Image: the image of this summoner spell
        """
        return Image(self.data.image) if self.data.image else None

    @property
    def key(self):
        """
        Returns:
            str: this summoner spell's identifying key
        """
        return self.data.key

    @cassiopeia.type.core.common.lazyproperty
    def leveltip(self):
        """
        Returns:
            LevelTip: the level-up tips for this spell
        """
        return LevelTip(self.data.leveltip) if self.data.leveltip else None

    @property
    def max_rank(self):
        """
        Returns:
            the maximum level for this spell
        """
        return self.data.maxrank

    @property
    def modes(self):
        """
        Returns:
            list<GameMode>: the game modes that this spell is allowed on
        """
        return [cassiopeia.type.core.common.GameMode(mode) for mode in self.data.modes]

    @property
    def name(self):
        """
        Returns:
            str: the name of this spell
        """
        return self.data.name

    @property
    def range(self):
        """
        Returns:
            list<int>: the level-by-level range of this spell
        """
        return self.data.range

    @property
    def range_burn(self):
        """
        Returns:
            str: formatted range of this spell by level
        """
        return self.data.rangeBurn

    @property
    def resource(self):
        """
        Returns:
            str: what resource this spell uses
        """
        return self.data.resource

    @property
    def sanitized_description(self):
        """
        Returns:
            str: the sanitized description of this summoner spell
        """
        return self.data.sanitizedDescription

    @property
    def sanitized_tooltip(self):
        """
        Returns:
            str: the sanitized tooltip for this summoner spell
        """
        return self.data.sanitizedTooltip

    @property
    def summoner_level(self):
        """
        Returns:
            int: the summoner level required to use this spell
        """
        return self.data.summonerLevel

    @property
    def tooltip(self):
        """
        Returns:
            str: the tooltip for this spell
        """
        return self.data.tooltip

    @cassiopeia.type.core.common.lazyproperty
    def variables(self):
        """
        Returns:
            SpellVariables: the variables that determine this spell's effects
        """
        return [SpellVariables(svars) for svars in self.data.vars]

    def __replace_variables(self, text, level):
        if level < 1 or level > 18:
            raise ValueError("Not a valid champion level")

        for svar in self.variables:
            val = svar.coefficients[level - 1] if svar.link == "@player.level" else svar.coefficients[0]
            text = text.replace("{{{{ {key} }}}}".format(key=svar.key), str(val))

        return text

    @cassiopeia.type.core.common.immutablemethod
    def tooltip_for_level(self, level):
        """Gets the tooltip for this spell for a specific level

        Args:
            level (int): the level of the champion

        Returns:
            str: the tooltip for that rank/level
        """
        return self.__replace_variables(self.tooltip, level)

    @cassiopeia.type.core.common.immutablemethod
    def sanitized_tooltip_for_level(self, level):
        """Gets the sanitized tooltip for this spell for a specific level

        Args:
            level (int): the level of the champion

        Returns:
            str: the sanitized tooltip for this summoner spell for the provided level
        """
        return self.__replace_variables(self.sanitized_tooltip, level)


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_rebind_all():
    SetItem.dto_type = cassiopeia.type.dto.staticdata.BlockItem
    ItemSet.dto_type = cassiopeia.type.dto.staticdata.Block
    SpellVariables.dto_type = cassiopeia.type.dto.staticdata.SpellVars
    LevelTip.dto_type = cassiopeia.type.dto.staticdata.LevelTip
    ChampionStats.dto_type = cassiopeia.type.dto.staticdata.Stats
    Skin.dto_type = cassiopeia.type.dto.staticdata.Skin
    RecommendedItems.dto_type = cassiopeia.type.dto.staticdata.Recommended
    Image.dto_type = cassiopeia.type.dto.staticdata.Image
    Passive.dto_type = cassiopeia.type.dto.staticdata.Passive
    ChampionInfo.dto_type = cassiopeia.type.dto.staticdata.Info
    Spell.dto_type = cassiopeia.type.dto.staticdata.ChampionSpell
    Champion.dto_type = cassiopeia.type.dto.staticdata.Champion
    MetaData.dto_type = cassiopeia.type.dto.staticdata.MetaData
    Gold.dto_type = cassiopeia.type.dto.staticdata.Gold
    ItemStats.dto_type = cassiopeia.type.dto.staticdata.BasicDataStats
    Item.dto_type = cassiopeia.type.dto.staticdata.Item
    MapDetails.dto_type = cassiopeia.type.dto.staticdata.MapDetails
    Mastery.dto_type = cassiopeia.type.dto.staticdata.Mastery
    Realm.dto_type = cassiopeia.type.dto.staticdata.Realm
    Rune.dto_type = cassiopeia.type.dto.staticdata.Rune
    SummonerSpell.dto_type = cassiopeia.type.dto.staticdata.SummonerSpell



import urllib.parse

import cassiopeia.dto.requests
import cassiopeia.type.dto.summoner


def get_summoners_by_name(summoner_names):
    """
    https://developer.riotgames.com/api/methods#!/1017/3446

    Args:
        summoner_names (str | list<str>): the summoner name(s) to look up

    Returns:
        dict<str, Summoner>: the requested summoners
    """
    # Can only have 40 summoners max if it's a list
    if isinstance(summoner_names, list) and len(summoner_names) > 40:
        raise ValueError("Can only get up to 40 summoners at once.")

    name_string = ",".join(urllib.parse.quote(x) for x in summoner_names) if isinstance(summoner_names, list) else urllib.parse.quote(summoner_names)

    # Get JSON response
    request = "{version}/summoner/by-name/{names}".format(version=cassiopeia.dto.requests.api_versions["summoner"], names=name_string)
    response = cassiopeia.dto.requests.get(request)

    # Convert response to Dto type
    for name, summoner in response.items():
        response[name] = cassiopeia.type.dto.summoner.Summoner(summoner)

    return response


def get_summoners_by_id(summoner_ids):
    """
    https://developer.riotgames.com/api/methods#!/1017/3447

    Args:
        summoner_ids (int | list<int>): the summoner ID(s) to look up

    Returns:
        dict<str, Summoner>: the requested summoners
    """
    # Can only have 40 summoners max if it's a list
    if isinstance(summoner_ids, list) and len(summoner_ids) > 40:
        raise ValueError("Can only get up to 40 summoners at once.")

    id_string = ",".join(str(x) for x in summoner_ids) if isinstance(summoner_ids, list) else str(summoner_ids)

    # Get JSON response
    request = "{version}/summoner/{ids}".format(version=cassiopeia.dto.requests.api_versions["summoner"], ids=id_string)
    response = cassiopeia.dto.requests.get(request)

    # Convert response to Dto type
    for id_, summoner in response.items():
        response[id_] = cassiopeia.type.dto.summoner.Summoner(summoner)

    return response


def get_summoner_masteries(summoner_ids):
    """
    https://developer.riotgames.com/api/methods#!/1017/3450

    Args:
        summoner_ids (int | list<int>): the summoner ID(s) to get mastery pages for

    Returns:
        dict<str, MasteryPages>: the requested summoners' mastery pages
    """
    # Can only have 40 summoners max if it's a list
    if isinstance(summoner_ids, list) and len(summoner_ids) > 40:
        raise ValueError("Can only get masteries for up to 40 summoners at once.")

    id_string = ",".join(str(x) for x in summoner_ids) if isinstance(summoner_ids, list) else str(summoner_ids)

    # Get JSON response
    request = "{version}/summoner/{ids}/masteries".format(version=cassiopeia.dto.requests.api_versions["summoner"], ids=id_string)
    response = cassiopeia.dto.requests.get(request)

    # Convert response to Dto type
    for id_, masteries in response.items():
        response[id_] = cassiopeia.type.dto.summoner.MasteryPages(masteries)

    return response


def get_summoner_names(summoner_ids):
    """
    https://developer.riotgames.com/api/methods#!/1017/3451

    Args:
        summoner_ids (int | list<int>): the summoner ID(s) to get names for

    Returns:
        dict<str, str>: the requested summoners' names
    """
    # Can only have 40 summoners max if it's a list
    if isinstance(summoner_ids, list) and len(summoner_ids) > 40:
        raise ValueError("Can only get names for up to 40 summoners at once.")

    id_string = ",".join(str(x) for x in summoner_ids) if isinstance(summoner_ids, list) else str(summoner_ids)

    # Get JSON response
    request = "{version}/summoner/{ids}/name".format(version=cassiopeia.dto.requests.api_versions["summoner"], ids=id_string)
    return cassiopeia.dto.requests.get(request)


def get_summoner_runes(summoner_ids):
    """
    https://developer.riotgames.com/api/methods#!/1017/3449

    Args:
        summoner_ids (int | list<int>): the summoner ID(s) to get rune pages for

    Returns:
        dict<str, RunePages>: the requested summoners' rune pages
    """
    # Can only have 40 summoners max if it's a list
    if isinstance(summoner_ids, list) and len(summoner_ids) > 40:
        raise ValueError("Can only get runes for up to 40 summoners at once.")

    id_string = ",".join(str(x) for x in summoner_ids) if isinstance(summoner_ids, list) else str(summoner_ids)

    # Get JSON response
    request = "{version}/summoner/{ids}/runes".format(version=cassiopeia.dto.requests.api_versions["summoner"], ids=id_string)
    response = cassiopeia.dto.requests.get(request)

    # Convert response to Dto type
    for id_, runes in response.items():
        response[id_] = cassiopeia.type.dto.summoner.RunePages(runes)

    return response



import cassiopeia.dto.gameapi
import cassiopeia.core.requests
import cassiopeia.type.core.game


def get_recent_games(summoner):
    """
    Gets the most recent games a summoner played

    Args:
        summoner (Summoner): the summoner to get recent games for

    Returns:
        list<Game>: the summoner's recent games
    """
    games = cassiopeia.dto.gameapi.get_recent_games(summoner.id)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        summoner_ids = games.summoner_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None
        cassiopeia.riotapi.get_summoner_spells() if games.summoner_spell_ids else None

    return [cassiopeia.type.core.game.Game(game, games.summonerId) for game in games.games]



import cassiopeia.dto.requests
import cassiopeia.type.dto.staticdata

_locale = None


def get_champion(id_):
    """
    https://developer.riotgames.com/api/methods#!/968/3322

    Args:
        id_ (int): the ID of the champion to get

    Returns:
        Champion: the champion
    """
    request = "{version}/champion/{id_}".format(version=cassiopeia.dto.requests.api_versions["staticdata"], id_=id_)

    params = {"champData": "all"}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.Champion(cassiopeia.dto.requests.get(request, params, True))


def get_champions():
    """
    https://developer.riotgames.com/api/methods#!/968/3326

    Returns:
        ChampionList: all the champions
    """
    request = "{version}/champion".format(version=cassiopeia.dto.requests.api_versions["staticdata"])

    params = {"champData": "all"}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.ChampionList(cassiopeia.dto.requests.get(request, params, True))


def get_item(id_):
    """
    https://developer.riotgames.com/api/methods#!/968/3319

    Args:
        id_ (int): the ID of the item to get

    Returns:
        Item: the item
    """
    request = "{version}/item/{id_}".format(version=cassiopeia.dto.requests.api_versions["staticdata"], id_=id_)

    params = {"itemData": "all"}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.Item(cassiopeia.dto.requests.get(request, params, True))


def get_items():
    """
    https://developer.riotgames.com/api/methods#!/968/3314

    Returns:
        ItemList: all the items
    """
    request = "{version}/item".format(version=cassiopeia.dto.requests.api_versions["staticdata"])

    params = {"itemListData": "all"}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.ItemList(cassiopeia.dto.requests.get(request, params, True))


def get_language_strings():
    """
    https://developer.riotgames.com/api/methods#!/968/3316

    Returns:
        LanguageStrings: the locale-based string replacements for various game constants
    """
    request = "{version}/language-strings".format(version=cassiopeia.dto.requests.api_versions["staticdata"])

    params = {}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.LanguageStrings(cassiopeia.dto.requests.get(request, params, True))


def get_languages():
    """
    https://developer.riotgames.com/api/methods#!/968/3324

    Returns:
        list<str>: the valid locale strings for the API
    """
    request = "{version}/languages".format(version=cassiopeia.dto.requests.api_versions["staticdata"])
    return cassiopeia.dto.requests.get(request, {}, True)


def get_maps():
    """
    https://developer.riotgames.com/api/methods#!/968/3328

    Returns:
        MapData: specific information about each map
    """
    request = "{version}/map".format(version=cassiopeia.dto.requests.api_versions["staticdata"])

    params = {}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.MapData(cassiopeia.dto.requests.get(request, params, True))


def get_mastery(id_):
    """
    https://developer.riotgames.com/api/methods#!/968/3318

    Args:
        id_ (int): the ID of the mastery to get

    Returns:
        Mastery: the mastery
    """
    request = "{version}/mastery/{id_}".format(version=cassiopeia.dto.requests.api_versions["staticdata"], id_=id_)

    params = {"masteryData": "all"}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.Mastery(cassiopeia.dto.requests.get(request, params, True))


def get_masteries():
    """
    https://developer.riotgames.com/api/methods#!/968/3317

    Returns:
        MasteryList: all the masteries
    """
    request = "{version}/mastery".format(version=cassiopeia.dto.requests.api_versions["staticdata"])

    params = {"masteryListData": "all"}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.MasteryList(cassiopeia.dto.requests.get(request, params, True))


def get_realm():
    """
    https://developer.riotgames.com/api/methods#!/968/3325

    Returns:
        Realm: the realm for the current region
    """
    request = "{version}/realm".format(version=cassiopeia.dto.requests.api_versions["staticdata"])
    return cassiopeia.type.dto.staticdata.Realm(cassiopeia.dto.requests.get(request, {}, True))


def get_rune(id_):
    """
    https://developer.riotgames.com/api/methods#!/968/3321

    Args:
        id_ (int): the ID of the rune to get

    Returns:
        Rune: the rune
    """
    request = "{version}/rune/{id_}".format(version=cassiopeia.dto.requests.api_versions["staticdata"], id_=id_)

    params = {"runeData": "all"}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.Rune(cassiopeia.dto.requests.get(request, params, True))


def get_runes():
    """
    https://developer.riotgames.com/api/methods#!/968/3315

    Returns:
        RuneList: all the runes
    """
    request = "{version}/rune".format(version=cassiopeia.dto.requests.api_versions["staticdata"])

    params = {"runeListData": "all"}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.RuneList(cassiopeia.dto.requests.get(request, params, True))


def get_summoner_spell(id_):
    """
    https://developer.riotgames.com/api/methods#!/968/3320

    Args:
        id_ (int): the ID of the summoner spell to get

    Returns:
        SummonerSpell: the summoner spell
    """
    request = "{version}/summoner-spell/{id_}".format(version=cassiopeia.dto.requests.api_versions["staticdata"], id_=id_)

    params = {"spellData": "all"}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.SummonerSpell(cassiopeia.dto.requests.get(request, params, True))


def get_summoner_spells():
    """
    https://developer.riotgames.com/api/methods#!/968/3327

    Returns:
        SummonerSpellList: all the summoner spells
    """
    request = "{version}/summoner-spell".format(version=cassiopeia.dto.requests.api_versions["staticdata"])

    params = {"spellData": "all"}
    if _locale:
        params["locale"] = _locale

    return cassiopeia.type.dto.staticdata.SummonerSpellList(cassiopeia.dto.requests.get(request, params, True))


def get_versions():
    """
    https://developer.riotgames.com/api/methods#!/968/3323

    Returns:
        list<str>: the valid API versions
    """
    request = "{version}/versions".format(version=cassiopeia.dto.requests.api_versions["staticdata"])
    return cassiopeia.dto.requests.get(request, {}, True)



import cassiopeia.type.dto.common
import cassiopeia.type.core.common


if cassiopeia.type.dto.common.sqlalchemy_imported:
    import sqlalchemy


@cassiopeia.type.core.common.inheritdocs
class Champion(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        active (bool): indicates if the champion is active
        botEnabled (bool): bot enabled flag (for custom games)
        botMmEnabled (bool): bot Match Made enabled flag (for Co-op vs. AI games)
        freeToPlay (bool): indicates if the champion is free to play. Free to play champions are rotated periodically
        id (int): champion ID. For static information correlating to champion IDs, please refer to the LoL Static Data API.
        rankedPlayEnabled (bool): ranked play enabled flag
    """
    def __init__(self, dictionary):
        self.active = dictionary.get("active", False)
        self.botEnabled = dictionary.get("botEnabled", False)
        self.botMmEnabled = dictionary.get("botMmEnabled", False)
        self.freeToPlay = dictionary.get("freeToPlay", False)
        self.id = dictionary.get("id", 0)
        self.rankedPlayEnabled = dictionary.get("rankedPlayEnabled", False)


@cassiopeia.type.core.common.inheritdocs
class ChampionList(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        champions (list<Champion>): the collection of champion information
    """
    def __init__(self, dictionary):
        self.champions = [(Champion(champ) if not isinstance(champ, Champion) else champ) for champ in dictionary.get("champions", []) if champ]

    @property
    def champion_ids(self):
        """
        Gets all champion IDs contained in this object
        """
        ids = set()
        for champ in self.champions:
            ids.add(champ.id)
        return ids


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_bind_champion():
    global Champion

    @cassiopeia.type.core.common.inheritdocs
    class Champion(Champion, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "ChampionStatus"
        active = sqlalchemy.Column(sqlalchemy.Boolean)
        botEnabled = sqlalchemy.Column(sqlalchemy.Boolean)
        botMmEnabled = sqlalchemy.Column(sqlalchemy.Boolean)
        freeToPlay = sqlalchemy.Column(sqlalchemy.Boolean)
        id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        rankedPlayEnabled = sqlalchemy.Column(sqlalchemy.Boolean)


def _sa_bind_all():
    _sa_bind_champion()



import cassiopeia.riotapi
import cassiopeia.dto.matchapi
import cassiopeia.core.requests
import cassiopeia.type.core.common
import cassiopeia.type.core.match
import cassiopeia.type.core.matchlist


def get_match(id_, include_timeline=True, tournament_code=""):
    """
    Gets a match

    Args:
        id_ (int | MatchReference): the ID of or reference to the match to get
        include_timeline (bool): whether to include timeline data in the returned match
        tournament_code (str): the tournament code if the match to be retrieved is from a tournament

    Returns:
        Match: the match
    """
    if isinstance(id_, cassiopeia.type.core.matchlist.MatchReference):
        id_ = id_.id

    match = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.match.Match, id_, "matchId")
    if match and (match.timeline or not include_timeline):
        return match

    match = cassiopeia.dto.matchapi.get_match(id_, include_timeline, tournament_code)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        cassiopeia.riotapi.get_items() if match.item_ids else None
        cassiopeia.riotapi.get_champions() if match.champion_ids else None
        cassiopeia.riotapi.get_masteries() if match.mastery_ids else None
        cassiopeia.riotapi.get_runes() if match.rune_ids else None
        summoner_ids = match.summoner_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None
        cassiopeia.riotapi.get_summoner_spells() if match.summoner_spell_ids else None

    match = cassiopeia.type.core.match.Match(match)
    cassiopeia.core.requests.data_store.store(match, id_)
    return match


def get_matches(ids, include_timeline=True, tournament_code=""):
    """
    Gets a bunch of matches

    Args:
        ids (list<int> | list<MatchReference>): the IDs of or references to the matches to get
        include_timeline (bool): whether to include timeline data in the returned matches
        tournament_code (str): the tournament code if the match to be retrieved is from a tournament

    Returns:
        list<Match>: the matches
    """
    ids = [ref.id if isinstance(ref, cassiopeia.type.core.matchlist.MatchReference) else ref for ref in ids]

    matches = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.match.Match, ids, "matchId")

    # Find which matches weren't cached
    missing = []
    loc = []
    for i in range(len(ids)):
        if not matches[i] or (include_timeline and not matches[i].timeline):
            missing.append(ids[i])
            loc.append(i)

    if not missing:
        return matches

    # Initialize eager loading variables appropriately
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        item_ids = set()
        champion_ids = set()
        mastery_ids = set()
        rune_ids = set()
        summoner_ids = set()
        summoner_spell_ids = set()

    # Make requests to get them
    for i in range(len(missing)):
        match = cassiopeia.type.core.match.Match(cassiopeia.dto.matchapi.get_match(missing[i], include_timeline, tournament_code))
        matches[loc[i]] = match
        missing[i] = match

        if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
            item_ids = item_ids | match.data.item_ids
            champion_ids = champion_ids | match.data.champion_ids
            mastery_ids = mastery_ids | match.data.mastery_ids
            rune_ids = rune_ids | match.data.rune_ids
            summoner_ids = summoner_ids | match.data.summoner_ids
            summoner_spell_ids = summoner_spell_ids | match.data.summoner_spell_ids

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        cassiopeia.riotapi.get_items() if item_ids else None
        cassiopeia.riotapi.get_champions() if champion_ids else None
        cassiopeia.riotapi.get_masteries() if mastery_ids else None
        cassiopeia.riotapi.get_runes() if rune_ids else None
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None
        cassiopeia.riotapi.get_summoner_spells() if summoner_spell_ids else None

    cassiopeia.core.requests.data_store.store(missing, [match.id for match in missing])
    return matches


def get_tournament_match_ids(tournament_code):
    """
    Gets the IDs for a tournament's matches

    Args:
        tournament_code (str): the tournament code

    Returns:
        list<int>: the match ids for the tournament
    """
    return cassiopeia.dto.matchapi.get_tournament_match_ids(tournament_code)



import cassiopeia.type.dto.common
import cassiopeia.type.core.common


if cassiopeia.type.dto.common.sqlalchemy_imported:
    import sqlalchemy
    import sqlalchemy.orm


@cassiopeia.type.core.common.inheritdocs
class ChampionMastery(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        championId (int): champion ID for this entry
        championLevel (int): champion level for specified player and champion combination
        championPoints (int): total number of champion points for this player and champion combination - they are used to determine championLevel
        championPointsSinceLastLevel (int): number of points earned since current level has been achieved. Zero if player reached maximum champion level for this champion.
        championPointsUntilNextLevel (int): number of points needed to achieve next level. Zero if player reached maximum champion level for this champion.
        chestGranted (bool): is chest granted for this champion or not in current season
        lastPlayTime (int): last time this champion was played by this player - in Unix milliseconds time format
        playerId (int): player ID for this entry
        tokensEarned (int): number of token earned for next level mastery
    """
    def __init__(self, dictionary):
        self.championId = dictionary.get("championId", 0)
        self.championLevel = dictionary.get("championLevel", 0)
        self.championPoints = dictionary.get("championPoints", 0)
        self.championPointsSinceLastLevel = dictionary.get("championPointsSinceLastLevel", 0)
        self.championPointsUntilNextLevel = dictionary.get("championPointsUntilNextLevel", 0)
        self.chestGranted = dictionary.get("chestGranted", False)
        self.lastPlayTime = dictionary.get("lastPlayTime", 0)
        self.playerId = dictionary.get("playerId", 0)
        self.tokensEarned = dictionary.get("tokensEarned", 0)


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_bind_champion_mastery():
    global ChampionMastery

    @cassiopeia.type.core.common.inheritdocs
    class ChampionMastery(ChampionMastery, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "ChampionMastery"
        championId = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        championLevel = sqlalchemy.Column(sqlalchemy.Integer)
        championPoints = sqlalchemy.Column(sqlalchemy.Integer)
        championPointsSinceLastLevel = sqlalchemy.Column(sqlalchemy.Integer)
        championPointsUntilNextLevel = sqlalchemy.Column(sqlalchemy.Integer)
        chestGranted = sqlalchemy.Column(sqlalchemy.Boolean)
        highestGrade = sqlalchemy.Column(sqlalchemy.String(30))
        lastPlayTime = sqlalchemy.Column(sqlalchemy.BigInteger)
        playerId = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)


def _sa_bind_all():
    _sa_bind_champion_mastery()



from cassiopeia import riotapi
from .. import int_test_handler


def test_all():
    print("core/summoner tests...")
    test_summoner_by_name()
    test_summoners_by_name()
    test_summoner_by_id()
    test_summoners_by_id()
    test_mastery_pages()
    test_summoner_name()
    test_summoner_names()
    test_rune_pages()


def test_summoner_by_name():
    int_test_handler.test_result(riotapi.get_summoner_by_name(int_test_handler.summoner_name))


def test_summoners_by_name():
    int_test_handler.test_result(riotapi.get_summoners_by_name([int_test_handler.summoner_name]))


def test_summoner_by_id():
    int_test_handler.test_result(riotapi.get_summoner_by_id(int_test_handler.summoner_id))


def test_summoners_by_id():
    int_test_handler.test_result(riotapi.get_summoners_by_id([int_test_handler.summoner_id]))


def test_mastery_pages():
    int_test_handler.test_result(riotapi.get_mastery_pages(riotapi.get_summoner_by_id(int_test_handler.summoner_id)))


def test_summoner_name():
    int_test_handler.test_result(riotapi.get_summoner_name(int_test_handler.summoner_id))


def test_summoner_names():
    int_test_handler.test_result(riotapi.get_summoner_names([int_test_handler.summoner_id]))


def test_rune_pages():
    int_test_handler.test_result(riotapi.get_rune_pages(riotapi.get_summoner_by_id(int_test_handler.summoner_id)))



import cassiopeia.riotapi
import cassiopeia.type.core.common
import cassiopeia.type.dto.champion


@cassiopeia.type.core.common.inheritdocs
class ChampionStatus(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.champion.Champion

    def __str__(self):
        return "Status ({champ})".format(champ=self.champion)

    @property
    def enabled(self):
        """
        Returns:
            bool: whether the champion is currently enabled
        """
        return self.data.active

    @property
    def custom_enabled(self):
        """
        Returns:
            bool: whether the champion is currently enabled for custom games
        """
        return self.data.botEnabled

    @property
    def coop_ai_enabled(self):
        """
        Returns:
            bool: whether the champion is currently enabled for coop vs ai games
        """
        return self.data.botMmEnabled

    @property
    def free(self):
        """
        Returns:
            bool: whether the champion is currently free this week
        """
        return self.data.freeToPlay

    @property
    def champion(self):
        """
        Returns:
            Champion: the Champion this status is for
        """
        return cassiopeia.riotapi.get_champion_by_id(self.data.id) if self.data.id else None

    @property
    def ranked_enabled(self):
        """
        Returns:
            bool: whether the champion is currently enabled for ranked games
        """
        return self.data.rankedPlayEnabled


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_rebind_all():
    ChampionStatus.dto_type = cassiopeia.type.dto.champion.Champion



from cassiopeia import riotapi
from .. import int_test_handler


def test_all():
    print("core/matchlistapi tests...")
    test_match_list()


def test_match_list():
    int_test_handler.test_result(riotapi.get_match_list(riotapi.get_summoner_by_id(int_test_handler.summoner_id)))



import cassiopeia.type.dto.common
import cassiopeia.type.core.common


if cassiopeia.type.dto.common.sqlalchemy_imported:
    import sqlalchemy
    import sqlalchemy.orm


@cassiopeia.type.core.common.inheritdocs
class Rune(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        count (int): the count of this rune used by the participant
        runeId (int): the ID of the rune
    """
    def __init__(self, dictionary):
        self.count = dictionary.get("count", 0)
        self.runeId = dictionary.get("runeId", 0)


@cassiopeia.type.core.common.inheritdocs
class Mastery(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        masteryId (int): the ID of the mastery
        rank (int): the number of points put into this mastery by the user
    """
    def __init__(self, dictionary):
        self.masteryId = dictionary.get("masteryId", 0)
        self.rank = dictionary.get("rank", 0)


@cassiopeia.type.core.common.inheritdocs
class Observer(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        encryptionKey (str): key used to decrypt the spectator grid game data for playback
    """
    def __init__(self, dictionary):
        self.encryptionKey = dictionary.get("encryptionKey", "")


@cassiopeia.type.core.common.inheritdocs
class CurrentGameParticipant(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        bot (bool): flag indicating whether or not this participant is a bot
        championId (int): the ID of the champion played by this participant
        masteries (list<Mastery>): the masteries used by this participant
        profileIconId (int): the ID of the profile icon used by this participant
        runes (list<Rune>): the runes used by this participant
        spell1Id (int): the ID of the first summoner spell used by this participant
        spell2Id (int): the ID of the second summoner spell used by this participant
        summonerId (int): the summoner ID of this participant
        summonerName (str): the summoner name of this participant
        teamId (int): the team ID of this participant, indicating the participant's team
    """
    def __init__(self, dictionary):
        self.bot = dictionary.get("bot", False)
        self.championId = dictionary.get("championId", 0)
        self.masteries = [(Mastery(mastery) if not isinstance(mastery, Mastery) else mastery) for mastery in dictionary.get("masteries", []) if mastery]
        self.profileIconId = dictionary.get("profileIconId", 0)
        self.runes = [(Rune(rune) if not isinstance(rune, Rune) else rune) for rune in dictionary.get("runes", []) if rune]
        self.spell1Id = dictionary.get("spell1Id", 0)
        self.spell2Id = dictionary.get("spell2Id", 0)
        self.summonerId = dictionary.get("summonerId", 0)
        self.summonerName = dictionary.get("summonerName", "")
        self.teamId = dictionary.get("teamId", 0)


@cassiopeia.type.core.common.inheritdocs
class BannedChampion(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        championId (int): the ID of the banned champion
        pickTurn (int): the turn during which the champion was banned
        teamId (int): the ID of the team that banned the champion
    """
    def __init__(self, dictionary):
        self.championId = dictionary.get("championId", 0)
        self.pickTurn = dictionary.get("pickTurn", 0)
        self.teamId = dictionary.get("teamId", 0)


@cassiopeia.type.core.common.inheritdocs
class CurrentGameInfo(cassiopeia.type.dto.common.CassiopeiaDto):
    """
    Args:
        bannedChampions (list<BannedChampion>): banned champion information
        gameId (int): the ID of the game
        gameLength (int): the amount of time in seconds that has passed since the game started
        gameMode (str): the game mode (Legal values: CLASSIC, ODIN, ARAM, TUTORIAL, ONEFORALL, ASCENSION, FIRSTBLOOD, KINGPORO)
        gameQueueConfigId (int): the queue type (queue types are documented on the Game Constants page)
        gameStartTime (int): the game start time represented in epoch milliseconds
        gameType (str): the game type (Legal values: CUSTOM_GAME, MATCHED_GAME, TUTORIAL_GAME)
        mapId (int): the ID of the map
        observers (Observer): the observer information
        participants (list<CurrentGameParticipant>): the participant information
        platformId (str): the ID of the platform on which the game is being played
    """
    def __init__(self, dictionary):
        self.bannedChampions = [(BannedChampion(ban) if not isinstance(ban, BannedChampion) else ban) for ban in dictionary.get("bannedChampions", []) if ban]
        self.gameId = dictionary.get("gameId", 0)
        self.gameLength = dictionary.get("gameLength", 0)
        self.gameMode = dictionary.get("gameMode", "")
        self.gameQueueConfigId = dictionary.get("gameQueueConfigId", 0)
        self.gameStartTime = dictionary.get("gameStartTime", 0)
        self.gameType = dictionary.get("gameType", "")
        self.mapId = dictionary.get("mapId", 0)
        val = dictionary.get("observers", None)
        self.observers = Observer(val) if val and not isinstance(val, Observer) else val
        self.participants = [(CurrentGameParticipant(participant) if not isinstance(participant, CurrentGameParticipant) else participant) for participant in dictionary.get("participants", []) if participant]
        self.platformId = dictionary.get("platformId", "")

    @property
    def champion_ids(self):
        """
        Gets all champion IDs contained in this object
        """
        ids = set()
        for ban in self.bannedChampions:
            ids.add(ban.championId)
        for p in self.participants:
            ids.add(p.championId)
        return ids

    @property
    def summoner_ids(self):
        """
        Gets all champion IDs contained in this object
        """
        ids = set()
        for p in self.participants:
            if p.summonerId:
                ids.add(p.summonerId)
        return ids

    @property
    def summoner_spell_ids(self):
        """
        Gets all champion IDs contained in this object
        """
        ids = set()
        for p in self.participants:
            if p.spell1Id:
                ids.add(p.spell1Id)
            if p.spell2Id:
                ids.add(p.spell2Id)
        return ids

    @property
    def rune_ids(self):
        """
        Gets all champion IDs contained in this object
        """
        ids = set()
        for p in self.participants:
            for r in p.runes:
                if r.runeId:
                    ids.add(r.runeId)
        return ids

    @property
    def mastery_ids(self):
        """
        Gets all champion IDs contained in this object
        """
        ids = set()
        for p in self.participants:
            for m in p.masteries:
                if m.masteryId:
                    ids.add(m.masteryId)
        return ids


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_bind_rune():
    global Rune

    @cassiopeia.type.core.common.inheritdocs
    class Rune(Rune, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "CurrentGameRune"
        count = sqlalchemy.Column(sqlalchemy.Integer)
        runeId = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _participant_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("CurrentGameParticipant._id", ondelete="CASCADE"))


def _sa_bind_mastery():
    global Mastery

    @cassiopeia.type.core.common.inheritdocs
    class Mastery(Mastery, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "CurrentGameMastery"
        masteryId = sqlalchemy.Column(sqlalchemy.Integer)
        rank = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _participant_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("CurrentGameParticipant._id", ondelete="CASCADE"))


def _sa_bind_observer():
    global Observer

    @cassiopeia.type.core.common.inheritdocs
    class Observer(Observer, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "CurrentGameObserver"
        encryptionKey = sqlalchemy.Column(sqlalchemy.String(50))
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _game_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("CurrentGameInfo.gameId", ondelete="CASCADE"))


def _sa_bind_current_game_participant():
    global CurrentGameParticipant

    @cassiopeia.type.core.common.inheritdocs
    class CurrentGameParticipant(CurrentGameParticipant, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "CurrentGameParticipant"
        bot = sqlalchemy.Column(sqlalchemy.Boolean)
        championId = sqlalchemy.Column(sqlalchemy.Integer)
        masteries = sqlalchemy.orm.relationship("cassiopeia.type.dto.currentgame.Mastery", cascade="all, delete-orphan", passive_deletes=True)
        profileIconId = sqlalchemy.Column(sqlalchemy.Integer)
        runes = sqlalchemy.orm.relationship("cassiopeia.type.dto.currentgame.Rune", cascade="all, delete-orphan", passive_deletes=True)
        spell1Id = sqlalchemy.Column(sqlalchemy.Integer)
        spell2Id = sqlalchemy.Column(sqlalchemy.Integer)
        summonerId = sqlalchemy.Column(sqlalchemy.Integer)
        summonerName = sqlalchemy.Column(sqlalchemy.String(30))
        teamId = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _game_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("CurrentGameInfo.gameId", ondelete="CASCADE"))


def _sa_bind_banned_champion():
    global BannedChampion

    @cassiopeia.type.core.common.inheritdocs
    class BannedChampion(BannedChampion, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "CurrentGameBannedChampion"
        championId = sqlalchemy.Column(sqlalchemy.Integer)
        pickTurn = sqlalchemy.Column(sqlalchemy.Integer)
        teamId = sqlalchemy.Column(sqlalchemy.Integer)
        _id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        _game_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("CurrentGameInfo.gameId", ondelete="CASCADE"))


def _sa_bind_current_game_info():
    global CurrentGameInfo

    @cassiopeia.type.core.common.inheritdocs
    class CurrentGameInfo(CurrentGameInfo, cassiopeia.type.dto.common.BaseDB):
        __tablename__ = "CurrentGameInfo"
        bannedChampions = sqlalchemy.orm.relationship("cassiopeia.type.dto.currentgame.BannedChampion", cascade="all, delete-orphan", passive_deletes=True)
        gameId = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
        gameLength = sqlalchemy.Column(sqlalchemy.Integer)
        gameMode = sqlalchemy.Column(sqlalchemy.String(30))
        gameQueueConfigId = sqlalchemy.Column(sqlalchemy.Integer)
        gameStartTime = sqlalchemy.Column(sqlalchemy.BigInteger)
        gameType = sqlalchemy.Column(sqlalchemy.String(30))
        mapId = sqlalchemy.Column(sqlalchemy.Integer)
        observers = sqlalchemy.orm.relationship("cassiopeia.type.dto.currentgame.Observer", uselist=False, cascade="all, delete-orphan", passive_deletes=True)
        participants = sqlalchemy.orm.relationship("cassiopeia.type.dto.currentgame.CurrentGameParticipant", cascade="all, delete-orphan", passive_deletes=True)
        platformId = sqlalchemy.Column(sqlalchemy.String(30))


def _sa_bind_all():
    _sa_bind_rune()
    _sa_bind_mastery()
    _sa_bind_observer()
    _sa_bind_current_game_participant()
    _sa_bind_banned_champion()
    _sa_bind_current_game_info()



import cassiopeia.riotapi
import cassiopeia.dto.teamapi
import cassiopeia.core.requests
import cassiopeia.type.core.team


def get_teams_by_summoner(summoners):
    """Gets (a) summoner(s)' teams

    summoners    Summoner | list<Summoner>        the summoner(s) to get teams for

    return       list<Team> | list<list<Team>>    the summoner(s)' teams
    """
    ids = [summoner.id for summoner in summoners] if isinstance(summoners, list) else summoners.id
    teams = cassiopeia.core.requests.call_with_ensured_size(cassiopeia.dto.teamapi.get_teams_by_summoner_id, 10, ids)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        summoner_ids = set()
        for _, lst in teams.items():
            for team in lst:
                summoner_ids |= team.summoner_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None

    if not isinstance(ids, list):
        return [cassiopeia.type.core.team.Team(team) for team in teams[str(ids)]]
    else:
        return [[cassiopeia.type.core.team.Team(team) for team in teams[str(id_)]] for id_ in ids]


def get_team(id_):
    """Gets a team by ID

    id_       str     the ID of the team

    return    Team    the team
    """
    team = cassiopeia.dto.teamapi.get_teams_by_id(id_)[id_]

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        summoner_ids = team.summoner_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None

    return cassiopeia.type.core.team.Team(team)


def get_teams(ids):
    """Gets teams by ID

    ids       list<str>     the IDs of the teams

    return    list<Team>    the teams
    """
    teams = cassiopeia.core.requests.call_with_ensured_size(cassiopeia.dto.teamapi.get_teams_by_id, 10, ids)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        summoner_ids = set()
        for _, team in teams.items():
            summoner_ids |= team.summoner_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None

    return [cassiopeia.type.core.team.Team(teams[id_]) for id_ in ids]



import cassiopeia.riotapi
import cassiopeia.dto.staticdataapi
import cassiopeia.core.requests
import cassiopeia.type.core.common
import cassiopeia.type.core.staticdata

_ignore_items = {0, 1080, 1304, 1309, 1314, 1319, 1324, 1329, 2004, 2037, 2039, 2040, 3005, 3039, 3123, 3128, 3131, 3160, 3166, 3167, 3168, 3169, 3175, 3176, 3171, 3186, 3188, 3205, 3206, 3207, 3209, 3210, 3244, 3405, 3406, 3407, 3408, 3409, 3410, 3411, 3412, 3413, 3414, 3415, 3416, 3417, 3419, 3420}
_ignore_runes = {8028}
_ignore_summoner_spells = {10}


def get_champion_by_id(id_):
    """
    Gets a champion by ID

    Args:
        id_ (int): the ID of the champion to get

    Returns:
        Champion: the champion
    """
    champion = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.staticdata.Champion, id_, "id")
    if champion:
        return champion

    champion = cassiopeia.dto.staticdataapi.get_champion(id_)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        cassiopeia.riotapi.get_items() if champion.item_ids else None

    champion = cassiopeia.type.core.staticdata.Champion(champion)
    cassiopeia.core.requests.data_store.store(champion, id_)
    return champion


def get_champion_by_name(name):
    """
    Gets a champion by name

    Args:
        name (str): the name of the champion to get

    Returns:
        Champion: the champion
    """
    champions = get_champions()
    for champion in champions:
        if champion.name == name:
            return champion
    return None


def get_champions():
    """
    Gets all the champions

    Returns:
        list<Champion>: all the champions
    """
    if cassiopeia.core.requests.data_store.has_all(cassiopeia.type.core.staticdata.Champion):
        return cassiopeia.core.requests.data_store.get_all(cassiopeia.type.core.staticdata.Champion)

    champions = cassiopeia.dto.staticdataapi.get_champions()

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        cassiopeia.riotapi.get_items() if champions.item_ids else None

    champions = [cassiopeia.type.core.staticdata.Champion(champ[1]) for champ in champions.data.items()]
    cassiopeia.core.requests.data_store.store(champions, [champ.id for champ in champions], [cassiopeia.type.core.staticdata.Champion])
    return champions


def get_champions_by_id(ids):
    """
    Gets a bunch of champions by ID

    Args:
        ids (list<int>): the IDs of the champions to get

    Returns:
        list<Champion>: the requested champions
    """
    champions = {champ.id: champ for champ in get_champions()}
    results = []
    for id_ in ids:
        try:
            champ = champions[id_]
        except KeyError:
            champ = None
        results.append(champ)
    return results


def get_champions_by_name(names):
    """
    Gets a bunch of champions by name

    Args:
        names (list<str>): the names of the champions to get

    Returns:
        list<Champion>: the requested champions
    """
    indices = {names[i]: i for i in range(len(names))}

    champions = get_champions()
    results = [None for _ in range(len(names))]
    for champion in champions:
        try:
            index = indices[champion.name]
            results[index] = champion
        except KeyError:
            pass

    return results


def get_item(id_):
    """
    Gets an item

    Args:
        id_ (int): the ID of the item to get

    Returns:
        Item: the item
    """
    if id_ in _ignore_items:
        return None

    item = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.staticdata.Item, id_, "id")
    if item:
        return item

    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        items = cassiopeia.riotapi.get_items()
        item = None
        for itm in items:
            if itm.id == id_:
                item = itm
                break
    else:
        item = cassiopeia.dto.staticdataapi.get_item(id_)
        item = cassiopeia.type.core.staticdata.Item(item)
        cassiopeia.core.requests.data_store.store(item, id_)

    return item


def get_items(ids=None):
    """
    Gets a bunch of items (or all of them)

    Args:
        ids (list<int>): the IDs of the items to get (or None to get all items) (default None)

    Returns:
        list<Item>: the items
    """
    if ids is not None:
        items = {item.id: item for item in get_items()}
        results = []
        for id_ in ids:
            try:
                item = items[id_]
            except KeyError:
                item = None
            results.append(item)
        return results
    else:
        if cassiopeia.core.requests.data_store.has_all(cassiopeia.type.core.staticdata.Item):
            return cassiopeia.core.requests.data_store.get_all(cassiopeia.type.core.staticdata.Item)

        items = cassiopeia.dto.staticdataapi.get_items()
        items = [cassiopeia.type.core.staticdata.Item(item[1]) for item in items.data.items()]

        cassiopeia.core.requests.data_store.store(items, [item.id for item in items], [cassiopeia.type.core.staticdata.Item])
        return items


def get_language_strings():
    """
    Gets the locale-based string replacements for various game constants

    Returns:
        return: dict<str, str>   the replacements
    """
    return cassiopeia.dto.staticdataapi.get_language_strings().data


def get_languages():
    """
    Gets the valid locales (languages) that can be used with the API

    Returns:
        list<str>: the valid locales
    """
    return cassiopeia.dto.staticdataapi.get_languages()


def get_map_information():
    """
    Gets specific information about each map

    Returns:
        list<MapDetails>: the map information
    """
    return [cassiopeia.type.core.staticdata.MapDetails(map_[1]) for map_ in cassiopeia.dto.staticdataapi.get_maps().data.items()]


def get_mastery(id_):
    """
    Gets a mastery

    Args:
        id_ (int): the ID of the mastery to get

    Returns:
        Mastery: the mastery
    """
    mastery = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.staticdata.Mastery, id_, "id")
    if mastery:
        return mastery

    mastery = cassiopeia.dto.staticdataapi.get_mastery(id_)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        cassiopeia.riotapi.get_masteries() if mastery.mastery_ids else None

    mastery = cassiopeia.type.core.staticdata.Mastery(mastery)

    cassiopeia.core.requests.data_store.store(mastery, id_)
    return mastery


def get_masteries(ids=None):
    """
    Gets a bunch of masteries (or all of them)

    Args:
        ids (list<int>): the IDs of the masteries to get (or None to get all masteries) (default None)

    Returns:
        list<Mastery>: the masteries
    """
    if ids is not None:
        masteries = {mastery.id: mastery for mastery in get_masteries()}
        results = []
        for id_ in ids:
            try:
                mastery = masteries[id_]
            except KeyError:
                mastery = None
            results.append(mastery)
        return results
    else:
        if cassiopeia.core.requests.data_store.has_all(cassiopeia.type.core.staticdata.Mastery):
            return cassiopeia.core.requests.data_store.get_all(cassiopeia.type.core.staticdata.Mastery)

        masteries = cassiopeia.dto.staticdataapi.get_masteries()
        masteries = [cassiopeia.type.core.staticdata.Mastery(mastery[1]) for mastery in masteries.data.items()]

        cassiopeia.core.requests.data_store.store(masteries, [mastery.id for mastery in masteries], [cassiopeia.type.core.staticdata.Mastery])
        return masteries


def get_realm():
    """
    Gets the realm for the current region

    Returns:
        Realm: the realm
    """
    return cassiopeia.type.core.staticdata.Realm(cassiopeia.dto.staticdataapi.get_realm())


def get_rune(id_):
    """
    Gets a rune

    Args:
        id_ (int): the ID of the rune to get

    Returns:
        Rune: the rune
    """
    if id_ in _ignore_runes:
        return None

    rune = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.staticdata.Rune, id_, "id")
    if rune:
        return rune

    rune = cassiopeia.dto.staticdataapi.get_rune(id_)
    rune = cassiopeia.type.core.staticdata.Rune(rune)

    cassiopeia.core.requests.data_store.store(rune, id_)
    return rune


def get_runes(ids=None):
    """
    Gets a bunch of runes (or all of them)

    Args:
        ids (list<int>): the IDs of the runes to get (or None to get all runes) (default None)

    Returns:
        list<Rune>: the runes
    """
    if ids is not None:
        runes = {rune.id: rune for rune in get_runes()}
        results = []
        for id_ in ids:
            try:
                rune = runes[id_]
            except KeyError:
                rune = None
            results.append(rune)
        return results
    else:
        if cassiopeia.core.requests.data_store.has_all(cassiopeia.type.core.staticdata.Rune):
            return cassiopeia.core.requests.data_store.get_all(cassiopeia.type.core.staticdata.Rune)

        runes = cassiopeia.dto.staticdataapi.get_runes()
        runes = [cassiopeia.type.core.staticdata.Rune(rune[1]) for rune in runes.data.items()]

        cassiopeia.core.requests.data_store.store(runes, [rune.id for rune in runes], [cassiopeia.type.core.staticdata.Rune])
        return runes


def get_summoner_spell(id_):
    """
    Gets a summoner spell

    Args:
        id_ (int): the ID of the summoner spell to get

    Returns:
        SummonerSpell: the summoner spell
    """
    if id_ in _ignore_summoner_spells:
        return None

    summoner_spell = cassiopeia.core.requests.data_store.get(cassiopeia.type.core.staticdata.SummonerSpell, id_, "id")
    if summoner_spell:
        return summoner_spell

    summoner_spell = cassiopeia.dto.staticdataapi.get_summoner_spell(id_)
    summoner_spell = cassiopeia.type.core.staticdata.SummonerSpell(summoner_spell)

    cassiopeia.core.requests.data_store.store(summoner_spell, id_)
    return summoner_spell


def get_summoner_spells(ids=None):
    """
    Gets a bunch of summoner spells (or all of them)

    Args:
        ids (list<int>): the IDs of the summoner spells to get (or None to get all summoner spells) (default None)

    Returns:
        list<SummonerSpell>: the summoner spells
    """
    if ids is not None:
        summoner_spells = {summoner_spell.id: summoner_spell for summoner_spell in get_summoner_spells()}
        results = []
        for id_ in ids:
            try:
                summoner_spell = summoner_spells[id_]
            except KeyError:
                summoner_spell = None
            results.append(summoner_spell)
        return results
    else:
        if cassiopeia.core.requests.data_store.has_all(cassiopeia.type.core.staticdata.SummonerSpell):
            return cassiopeia.core.requests.data_store.get_all(cassiopeia.type.core.staticdata.SummonerSpell)

        summoner_spells = cassiopeia.dto.staticdataapi.get_summoner_spells()
        summoner_spells = [cassiopeia.type.core.staticdata.SummonerSpell(summoner_spell[1]) for summoner_spell in summoner_spells.data.items()]

        cassiopeia.core.requests.data_store.store(summoner_spells, [summoner_spell.id for summoner_spell in summoner_spells], [cassiopeia.type.core.staticdata.SummonerSpell])
        return summoner_spells


def get_versions():
    """
    Gets the valid versions of the API

    Returns:
        list<str>: the valid versions
    """
    return cassiopeia.dto.staticdataapi.get_versions()



from cassiopeia import riotapi
from .. import int_test_handler


def test_all():
    print("core/leagueapi tests...")
    test_leagues_by_summoner()
    test_league_entries_by_summoner()
    test_leagues_by_team()
    test_league_entries_by_team()
    test_challenger()
    test_master()


def test_leagues_by_summoner():
    int_test_handler.test_result(riotapi.get_leagues_by_summoner(riotapi.get_summoner_by_id(int_test_handler.summoner_id)))


def test_league_entries_by_summoner():
    int_test_handler.test_result(riotapi.get_league_entries_by_summoner(riotapi.get_summoner_by_id(int_test_handler.summoner_id)))


def test_leagues_by_team():
    int_test_handler.test_result(riotapi.get_leagues_by_team(riotapi.get_team(int_test_handler.team_id)))


def test_league_entries_by_team():
    int_test_handler.test_result(riotapi.get_league_entries_by_team(riotapi.get_team(int_test_handler.team_id)))


def test_challenger():
    int_test_handler.test_result(riotapi.get_challenger())


def test_master():
    int_test_handler.test_result(riotapi.get_master())



import cassiopeia.dto.requests
import cassiopeia.type.dto.status


def get_shards():
    """
    https://developer.riotgames.com/api/methods#!/908/3143

    Returns:
        list<Shard>: the shards (unfortunately neither Crystal nor Kirby's)
    """
    request = "http://status.leagueoflegends.com/shards"
    return [cassiopeia.type.dto.status.Shard(shard) for shard in cassiopeia.dto.requests.get(request, static=True, include_base=False)]


def get_shard():
    """
    https://developer.riotgames.com/api/methods#!/908/3142

    Returns:
        ShardStatus: the status of the current region's shard
    """
    request = "http://status.leagueoflegends.com/shards/{region}".format(region=cassiopeia.dto.requests.region)
    return cassiopeia.type.dto.status.ShardStatus(cassiopeia.dto.requests.get(request, static=True, include_base=False))



"""
Handles making HTTP requests to the REST API and converting the results into a usable format
"""

import urllib.parse
import urllib.request
import urllib.error
import json
import zlib
import time

import cassiopeia.type.api.exception
import cassiopeia.type.api.rates


api_versions = {
    "champion": "v1.2",
    "championmastery": None,
    "currentgame": "v1.0",
    "featuredgames": "v1.0",
    "game": "v1.3",
    "league": "v2.5",
    "staticdata": "v1.2",
    "status": "v1.0",
    "match": "v2.2",
    "matchlist": "v2.2",
    "stats": "v1.3",
    "summoner": "v1.4",
    "team": "v2.4",
    "tournament": "v1"
}

api_key = ""
tournament_api_key = ""
region = ""
print_calls = False
rate_limiter = None
tournament_rate_limiter = None


def get(request, params={}, static=False, include_base=True, tournament=False):
    return make_request(request=request, method="GET", params=params, static=static, include_base=include_base, tournament=tournament)


def put(request, payload, params={}, include_base=True, tournament=False):
    return make_request(request=request, method="PUT", payload=payload, params=params, include_base=include_base, tournament=tournament)


def post(request, payload, params={}, include_base=True, tournament=False):
    return make_request(request=request, method="POST", payload=payload, params=params, include_base=include_base, tournament=tournament)


def make_request(request, method, params={}, payload=None, static=False, include_base=True, tournament=False):
    """
    Makes a rate-limited HTTP request to the Riot API and returns the result

    Args:
        request (str): the request string
        method (str): the HTTP method to use
        params (dict<str, any>): the path parameters to send with the request (default {})
        payload (CassiopeiaDto | CassiopeiaObject): the payload to send with the POST or PUT request (default None)
        static (bool): whether this is a call to a static (non-rate-limited) API (default False)
        include_base (bool): whether to prepend https://{server}.api.pvp.net/api/lol/{region}/ to the request (default True)
        tournament (bool): whether to use the tournament API rate limit (default False)

    Returns:
        dict: the JSON response from the Riot API as a dict
    """
    if (not tournament and not api_key) or (tournament and not tournament_api_key):
        raise cassiopeia.type.api.exception.CassiopeiaException("API Key must be set before the API can be queried.")
    if not region:
        raise cassiopeia.type.api.exception.CassiopeiaException("Region must be set before the API can be queried.")

    # Set server and rgn
    server = "global" if static else region
    rgn = ("static-data/{region}" if static else "{region}").format(region=region)

    # Encode params
    params["api_key"] = tournament_api_key if tournament else api_key
    encoded_params = urllib.parse.urlencode(params)

    payload = payload.to_json(separators=(",", ":"), indent=None) if payload else ""

    # Build and execute request
    if include_base:
        url = "https://{server}.api.pvp.net/api/lol/{region}/{request}?{params}".format(server=server, region=rgn, request=request, params=encoded_params)
    else:
        url = "{request}?{params}".format(request=request, params=encoded_params)

    limiter = tournament_rate_limiter if tournament else rate_limiter
    try:
        content = limiter.call(execute_request, url, method, payload) if limiter else execute_request(url, method, payload)
        return json.loads(content) if content else {}
    except urllib.error.HTTPError as e:
        # Reset rate limiter and retry on 429 (rate limit exceeded)
        if e.code == 429 and limiter:
            if "X-Rate-Limit-Type" not in e.headers or e.headers["X-Rate-Limit-Type"] == "service":
                time.sleep(1)  # Backoff for 1 second before retrying
            else:
                retry_after = 1
                if e.headers["Retry-After"]:
                    retry_after += int(e.headers["Retry-After"])

                limiter.reset_in(retry_after)
            return make_request(request, method, params, payload, static, include_base, tournament)
        else:
            raise cassiopeia.type.api.exception.APIError("Server returned error {code} on call: {url}".format(code=e.code, url=url), e.code)


def execute_request(url, method, payload=""):
    """
    Executes an HTTP request and returns the result in a string

    Args:
        url (str): the full URL to send a request to
        method (str): the HTTP method to use
        payload (str): the json payload to send if appropriate for HTTP method (default "")

    Returns:
        str: the content returned by the server
    """
    if print_calls:
        print(url)

    response = None
    try:
        if payload:
            payload = payload.encode("UTF-8")
            request = urllib.request.Request(url, method=method, data=payload)
            request.add_header("Content-Type", "application/json")
        else:
            request = urllib.request.Request(url, method=method)
        request.add_header("Accept-Encoding", "gzip")
        response = urllib.request.urlopen(request)
        content = response.read()
        if content:
            if "gzip" == response.getheader("Content-Encoding"):
                content = zlib.decompress(content, zlib.MAX_WBITS | 16).decode(encoding="UTF-8")
            else:
                content = content.decode("UTF-8")
        return content
    finally:
        if response:
            response.close()



import cassiopeia.dto.requests
import cassiopeia.type.dto.league


def get_leagues_by_summoner(summoner_ids):
    """
    https://developer.riotgames.com/api/methods#!/985/3351

    Args:
        summoner_ids (int | list<int>): the summoner ID(s) to get leagues for

    Returns:
        dict<str, list<League>>: the summoner(s)' leagues
    """
    # Can only have 10 summoners max if it's a list
    if isinstance(summoner_ids, list) and len(summoner_ids) > 10:
        raise ValueError("Can only get leagues for up to 10 summoners at once.")

    id_string = ",".join(str(x) for x in summoner_ids) if isinstance(summoner_ids, list) else str(summoner_ids)

    # Get JSON response
    request = "{version}/league/by-summoner/{ids}".format(version=cassiopeia.dto.requests.api_versions["league"], ids=id_string)
    response = cassiopeia.dto.requests.get(request)

    # Convert response to Dto type
    for id_, leagues in response.items():
        for i, league in enumerate(leagues):
            leagues[i] = cassiopeia.type.dto.league.League(league)

    return response


def get_league_entries_by_summoner(summoner_ids):
    """
    https://developer.riotgames.com/api/methods#!/985/3356

    Args:
        summoner_ids (int | list<int>): the summoner ID(s) to get league entries for

    Returns:
        dict<str, list<League>>: the summoner(s)' league entries
    """
    # Can only have 10 summoners max if it's a list
    if isinstance(summoner_ids, list) and len(summoner_ids) > 10:
        raise ValueError("Can only get leagues for up to 10 summoners at once.")

    id_string = ",".join(str(x) for x in summoner_ids) if isinstance(summoner_ids, list) else str(summoner_ids)

    # Get JSON response
    request = "{version}/league/by-summoner/{ids}/entry".format(version=cassiopeia.dto.requests.api_versions["league"], ids=id_string)
    response = cassiopeia.dto.requests.get(request)

    # Convert response to Dto type
    for id_, leagues in response.items():
        for i, league in enumerate(leagues):
            leagues[i] = cassiopeia.type.dto.league.League(league)

    return response


def get_leagues_by_team(team_ids):
    """
    https://developer.riotgames.com/api/methods#!/985/3352

    Args:
        team_ids (str | list<str>): the team ID(s) to get leagues for

    Returns:
        dict<str, list<League>>: the team(s)' leagues
    """
    # Can only have 10 teams max if it's a list
    if isinstance(team_ids, list) and len(team_ids) > 10:
        raise ValueError("Can only get leagues for up to 10 teams at once.")

    id_string = ",".join(team_ids) if isinstance(team_ids, list) else str(team_ids)

    # Get JSON response
    request = "{version}/league/by-team/{ids}".format(version=cassiopeia.dto.requests.api_versions["league"], ids=id_string)
    response = cassiopeia.dto.requests.get(request)

    # Convert response to Dto type
    for id_, leagues in response.items():
        response[id_] = [cassiopeia.type.dto.league.League(league) for league in leagues]

    return response


def get_league_entries_by_team(team_ids):
    """
    https://developer.riotgames.com/api/methods#!/985/3355

    Args:
        team_ids (str | list<str>): the team ID(s) to get league entries for

    Returns:
        dict<str, list<League>>: the team(s)' league entries
    """
    # Can only have 10 teams max if it's a list
    if isinstance(team_ids, list) and len(team_ids) > 10:
        raise ValueError("Can only get league entries for up to 10 teams at once.")

    id_string = ",".join(team_ids) if isinstance(team_ids, list) else str(team_ids)

    # Get JSON response
    request = "{version}/league/by-team/{ids}/entry".format(version=cassiopeia.dto.requests.api_versions["league"], ids=id_string)
    response = cassiopeia.dto.requests.get(request)

    # Convert response to Dto type
    for id_, leagues in response.items():
        response[id_] = [cassiopeia.type.dto.league.League(league) for league in leagues]

    return response


def get_challenger(queue_type):
    """
    https://developer.riotgames.com/api/methods#!/985/3353

    Args:
        queue_type (str): the queue type to get the challenger league for ("RANKED_SOLO_5x5", "RANKED_TEAM_3x3", "RANKED_TEAM_5x5")

    Returns:
        League: the challenger league
    """
    request = "{version}/league/challenger".format(version=cassiopeia.dto.requests.api_versions["league"])
    return cassiopeia.type.dto.league.League(cassiopeia.dto.requests.get(request, {"type": queue_type}))


def get_master(queue_type):
    """
    https://developer.riotgames.com/api/methods#!/985/3354

    Args:
        queue_type (str): the queue type to get the master league for ("RANKED_SOLO_5x5", "RANKED_TEAM_3x3", "RANKED_TEAM_5x5")

    Returns:
        League: the master league
    """
    request = "{version}/league/master".format(version=cassiopeia.dto.requests.api_versions["league"])
    return cassiopeia.type.dto.league.League(cassiopeia.dto.requests.get(request, {"type": queue_type}))



import cassiopeia.dto.requests
import cassiopeia.type.dto.featuredgames


def get_featured_games():
    """
    https://developer.riotgames.com/api/methods#!/977/3337

    Returns:
        FeaturedGames: the current featured game list
    """
    request = "https://{server}.api.pvp.net/observer-mode/rest/featured".format(server=cassiopeia.dto.requests.region)
    return cassiopeia.type.dto.featuredgames.FeaturedGames(cassiopeia.dto.requests.get(request, include_base=False))



from cassiopeia import baseriotapi
from .. import int_test_handler


def test_all():
    print("dto/stats tests...")
    test_ranked_stats()
    test_stats()


def test_ranked_stats():
    int_test_handler.test_result(baseriotapi.get_ranked_stats(int_test_handler.summoner_id))


def test_stats():
    int_test_handler.test_result(baseriotapi.get_stats(int_test_handler.summoner_id))



from cassiopeia import baseriotapi
from .. import int_test_handler


def test_all():
    print("dto/summoner tests...")
    test_summoners_by_name()
    test_summoners_by_id()
    test_summoner_masteries()
    test_summoner_names()
    test_summoner_runes()


def test_summoners_by_name():
    int_test_handler.test_result(baseriotapi.get_summoners_by_name(int_test_handler.summoner_name))


def test_summoners_by_id():
    int_test_handler.test_result(baseriotapi.get_summoners_by_id(int_test_handler.summoner_id))


def test_summoner_masteries():
    int_test_handler.test_result(baseriotapi.get_summoner_masteries(int_test_handler.summoner_id))


def test_summoner_names():
    int_test_handler.test_result(baseriotapi.get_summoner_names(int_test_handler.summoner_id))


def test_summoner_runes():
    int_test_handler.test_result(baseriotapi.get_summoner_runes(int_test_handler.summoner_id))



import datetime

import cassiopeia.type.core.common
import cassiopeia.type.dto.status


@cassiopeia.type.core.common.inheritdocs
class Shard(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.status.Shard

    def __str__(self):
        return self.name

    @property
    def host_name(self):
        """
        Returns:
            str: the domain name of the server
        """
        return self.data.hostname

    @property
    def locales(self):
        """
        Returns:
            list<str>: the languages that you can have api results in
        """
        return self.data.locales

    @property
    def name(self):
        """
        Returns:
            str: the name of this service
        """
        return self.data.name

    @property
    def platform(self):
        """
        Returns:
            Platform: the platform (i.e. server) for this match
        """
        return cassiopeia.type.core.common.Platform(self.data.region_tag.upper()) if self.data.region_tag else None

    @property
    def region(self):
        """
        Returns:
            Region: the region of the server is located in
        """
        return cassiopeia.type.core.common.Region(self.data.slug.upper()) if self.data.slug else None


@cassiopeia.type.core.common.inheritdocs
class ShardStatus(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.status.ShardStatus

    def __str__(self):
        return "Status ({name})".format(name=self.name)

    def __iter__(self):
        return iter(self.services)

    def __len__(self):
        return len(self.services)

    def __getitem__(self, index):
        return self.services[index]

    @property
    def host_name(self):
        """
        Returns:
            str: the domain name of the server
        """
        return self.data.hostname

    @property
    def locales(self):
        """
        Returns:
            list<str>: the languages that you can have api results in
        """
        return self.data.locales

    @property
    def name(self):
        """
        Returns:
            str: the full name of the region the server is located in
        """
        return self.data.name

    @property
    def platform(self):
        """
        Returns:
            Platform: the platform (i.e. server) for this match
        """
        return cassiopeia.type.core.common.Platform(self.data.region_tag.upper()) if self.data.region_tag else None

    @cassiopeia.type.core.common.lazyproperty
    def services(self):
        """
        Returns:
            list<Service>: the services that this region offers
        """
        return [Service(service) for service in self.data.services]

    @property
    def region(self):
        """
        Returns:
            Region: the region of the server is located in
        """
        return cassiopeia.type.core.common.Region(self.data.slug.upper()) if self.data.slug else None


@cassiopeia.type.core.common.inheritdocs
class Service(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.status.Service

    def __str__(self):
        return self.name

    def __iter__(self):
        return iter(self.incidents)

    def __len__(self):
        return len(self.incidents)

    def __getitem__(self, index):
        return self.incidents[index]

    @cassiopeia.type.core.common.lazyproperty
    def incidents(self):
        """
        Returns:
            list<Incident>: the incidents associated with this server
        """
        return [Incident(incident) for incident in self.data.incidents]

    @property
    def name(self):
        """
        Returns:
            str: the name of this service
        """
        return self.data.name

    @property
    def slug(self):
        """
        Returns:
            str: the name of the service in lowercase
        """
        return self.data.slug

    @property
    def status(self):
        """
        Returns:
            str: the status of the service
        """
        return self.data.status


@cassiopeia.type.core.common.inheritdocs
class Incident(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.status.Incident

    def __str__(self):
        return "Incident #{id_}".format(id_=self.id)

    def __iter__(self):
        return iter(self.updates)

    def __len__(self):
        return len(self.updates)

    def __getitem__(self, index):
        return self.updates[index]

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def active(self):
        """
        Returns:
            boolean: whether or not this incident is active
        """
        return self.data.active

    @cassiopeia.type.core.common.lazyproperty
    def created(self):
        """
        Returns:
            datetime.datetime: whent his message was created
        """
        return datetime.datetime.strptime(self.data.created_at, "%Y-%m-%dT%H:%M:%SZ") if self.data.created_at else None

    @property
    def id(self):
        """
        Returns:
            int: the id of this message
        """
        return self.data.id

    @cassiopeia.type.core.common.lazyproperty
    def updates(self):
        """
        Returns:
            list<Message>: the updates associated with this incident
        """
        return [Message(update) for update in self.data.updates]


@cassiopeia.type.core.common.inheritdocs
class Message(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.status.Message

    def __str__(self):
        return self.content

    def __iter__(self):
        return iter(self.translations)

    def __len__(self):
        return len(self.translations)

    def __getitem__(self, index):
        return self.translations[index]

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def author(self):
        """
        Returns:
            str: who wrote this message
        """
        return self.author

    @property
    def content(self):
        """
        Returns:
            str: the content of this message
        """
        return self.content

    @cassiopeia.type.core.common.lazyproperty
    def created(self):
        """
        Returns:
            datetime.datetime: whent his incident was created
        """
        return datetime.datetime.strptime(self.data.created_at, "%Y-%m-%dT%H:%M:%SZ") if self.data.created_at else None

    @property
    def id(self):
        """
        Returns:
            int: the id of this incident
        """
        return self.data.id

    @property
    def severity(self):
        """
        Returns:
            str: the severity of this message
        """
        return self.data.severity

    @cassiopeia.type.core.common.lazyproperty
    def translations(self):
        """
        Returns:
            dict<translation.locale: Translation>: the translated text of this message
        """
        return {translation.locale: Translation(translation) for translation in self.data.translations}

    @cassiopeia.type.core.common.lazyproperty
    def updated(self):
        """
        Returns:
            datetime.datetime: when this message was last updated
        """
        return datetime.datetime.strptime(self.data.updated_at, "%Y-%m-%dT%H:%M:%SZ") if self.data.updated_at else None


@cassiopeia.type.core.common.inheritdocs
class Translation(cassiopeia.type.core.common.CassiopeiaObject):
    dto_type = cassiopeia.type.dto.status.Translation

    def __str__(self):
        return self.content

    @property
    def content(self):
        """
        Returns:
            str: the content of this translation
        """
        return self.data.content

    @property
    def locale(self):
        """
        Returns:
            str: the content of this translation
        """
        return self.data.locale

    @cassiopeia.type.core.common.lazyproperty
    def updated(self):
        """
        Returns:
            datetime.datetime: when this translation was last updated
        """
        return datetime.datetime.strptime(self.data.updated_at, "%Y-%m-%dT%H:%M:%SZ") if self.data.updated_at else None


###############################
# Dynamic SQLAlchemy bindings #
###############################
def _sa_rebind_all():
    Shard.dto_type = cassiopeia.type.dto.status.Shard
    ShardStatus.dto_type = cassiopeia.type.dto.status.ShardStatus
    Service.dto_type = cassiopeia.type.dto.status.Service
    Incident.dto_type = cassiopeia.type.dto.status.Incident
    Message.dto_type = cassiopeia.type.dto.status.Message
    Translation.dto_type = cassiopeia.type.dto.status.Translation



import cassiopeia.riotapi
import cassiopeia.dto.leagueapi
import cassiopeia.core.requests
import cassiopeia.type.core.common
import cassiopeia.type.core.league


def get_challenger(queue_type=cassiopeia.type.core.common.Queue.ranked_solo):
    """
    Gets the challenger league

    Args:
        queue_type (Queue): the queue to get the challenger league for (default Queue.ranked_solo)

    Returns:
        League: the challenger league for that queue
    """
    if queue_type not in cassiopeia.type.core.common.ranked_queues:
        raise ValueError("Must use a ranked queue type to get ranked leagues")

    league = cassiopeia.dto.leagueapi.get_challenger(queue_type.value)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        summoner_ids = league.summoner_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None
        team_ids = league.team_ids
        cassiopeia.riotapi.get_teams(list(team_ids)) if team_ids else None

    return cassiopeia.type.core.league.League(league)


def get_master(queue_type=cassiopeia.type.core.common.Queue.ranked_solo):
    """
    Gets the master league

    Args:
        queue_type (Queue): the queue to get the master league for (default Queue.ranked_solo)

    Returns:
        League: the master league for that queue
    """
    if queue_type not in cassiopeia.type.core.common.ranked_queues:
        raise ValueError("Must use a ranked queue type to get ranked leagues")

    league = cassiopeia.dto.leagueapi.get_master(queue_type.value)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        summoner_ids = league.summoner_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None
        team_ids = league.team_ids
        cassiopeia.riotapi.get_teams(list(team_ids)) if team_ids else None

    return cassiopeia.type.core.league.League(league)


def __get_leagues_by_summoner_id(ids):
    leagues = cassiopeia.core.requests.call_with_ensured_size(cassiopeia.dto.leagueapi.get_leagues_by_summoner, 10, ids)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        summoner_ids = set()
        team_ids = set()
        for summoner in leagues.items():
            for league in summoner[1]:
                summoner_ids = summoner_ids | league.summoner_ids
                team_ids = team_ids | league.team_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None
        cassiopeia.riotapi.get_teams(list(team_ids)) if team_ids else None

    if not isinstance(ids, list):
        return [cassiopeia.type.core.league.League(league) for league in leagues[str(ids)]]
    else:
        return [[cassiopeia.type.core.league.League(league) for league in leagues.get(str(id_), []) if leagues.get(str(id_), [])] for id_ in ids]


def get_leagues_by_summoner(summoners):
    """
    Gets the leagues that the summoner(s) belong(s) to. You probably don't want to call this with LoadPolicy.eager set.

    Args:
        summoners (Summoner | list<Summoner>): the summoner(s) to get leagues for

    Returns:
        list<League> | list<list<League>>: the leagues that the requested summoner(s) belong(s) to
    """
    if isinstance(summoners, list):
        return __get_leagues_by_summoner_id([summoner.id for summoner in summoners])
    else:
        return __get_leagues_by_summoner_id(summoners.id)


def __get_league_entries_by_summoner_id(ids):
    leagues = cassiopeia.core.requests.call_with_ensured_size(cassiopeia.dto.leagueapi.get_league_entries_by_summoner, 10, ids)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        summoner_ids = set()
        team_ids = set()
        for summoner in leagues.items():
            for league in summoner[1]:
                summoner_ids = summoner_ids | league.summoner_ids
                team_ids = team_ids | league.team_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None
        cassiopeia.riotapi.get_teams(list(team_ids)) if team_ids else None

    if not isinstance(ids, list):
        return [cassiopeia.type.core.league.League(league) for league in leagues[str(ids)]]
    else:
        return [[cassiopeia.type.core.league.League(league) for league in leagues.get(str(id_), []) if leagues.get(str(id_), [])] for id_ in ids]


def get_league_entries_by_summoner(summoners):
    """
    Gets the leagues that the summoner(s) belong(s) to, including only the requested summoner(s)' entries

    Args:
        summoners (Summoner | list<Summoner>): the summoner(s) to get leagues for

    Returns:
        list<League> | list<list<League>>: the leagues that the requested summoner(s) belong(s) to
    """
    if isinstance(summoners, list):
        return __get_league_entries_by_summoner_id([summoner.id for summoner in summoners])
    else:
        return __get_league_entries_by_summoner_id(summoners.id)


def __get_leagues_by_team_id(ids):
    leagues = cassiopeia.core.requests.call_with_ensured_size(cassiopeia.dto.leagueapi.get_leagues_by_team, 10, ids)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        summoner_ids = set()
        team_ids = set()
        for team in leagues.items():
            for league in team[1]:
                summoner_ids = summoner_ids | league.summoner_ids
                team_ids = team_ids | league.team_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None
        cassiopeia.riotapi.get_teams(list(team_ids)) if team_ids else None

    if not isinstance(ids, list):
        return [cassiopeia.type.core.league.League(league) for league in leagues[ids]]
    else:
        return [[cassiopeia.type.core.league.League(league) for league in leagues.get(id_, []) if leagues.get(id_, [])] for id_ in ids]


def get_leagues_by_team(teams):
    """
    Gets the leagues that the team(s) belong(s) to. You probably don't want to call this with LoadPolicy.eager set.

    Args:
        teams (Team | list<Team>): the team(s) to get leagues for

    Returns:
        list<League> | list<list<League>>: the leagues that the requested team(s) belong(s) to
    """
    if isinstance(teams, list):
        return __get_leagues_by_team_id([team.id for team in teams])
    else:
        return __get_leagues_by_team_id(teams.id)


def __get_league_entries_by_team_id(ids):
    leagues = cassiopeia.core.requests.call_with_ensured_size(cassiopeia.dto.leagueapi.get_league_entries_by_team, 10, ids)

    # Load required data if loading policy is eager
    if cassiopeia.core.requests.load_policy is cassiopeia.type.core.common.LoadPolicy.eager:
        summoner_ids = set()
        team_ids = set()
        for team in leagues.items():
            for league in team[1]:
                summoner_ids = summoner_ids | league.summoner_ids
                team_ids = team_ids | league.team_ids
        cassiopeia.riotapi.get_summoners_by_id(list(summoner_ids)) if summoner_ids else None
        cassiopeia.riotapi.get_teams(list(team_ids)) if team_ids else None

    if not isinstance(ids, list):
        return [cassiopeia.type.core.league.League(league) for league in leagues[ids]]
    else:
        return [[cassiopeia.type.core.league.League(league) for league in leagues.get(id_, []) if leagues.get(id_, [])] for id_ in ids]


def get_league_entries_by_team(teams):
    """
    Gets the leagues that the team(s) belong(s) to, including only the requested team(s)' entries

    Args:
        teams (Team | list<Team>): the team(s) to get leagues for

    Returns:
        list<League> | list<list<League>>: the leagues that the requested team(s) belong(s) to
    """
    if isinstance(teams, list):
        return __get_league_entries_by_team_id([team.id for team in teams])
    else:
        return __get_league_entries_by_team_id(teams.id)
